# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
from pyomo.repn.standard_repn import generate_standard_repn
from mpi4py import MPI
from mpisppy.utils.lshaped_cuts import LShapedCutGenerator

import numpy as np
import pyomo.environ as pyo
import mpisppy.cylinders.spoke as spoke

class CrossScenarioCutSpoke(spoke.Spoke):
    def __init__(self, spbase_object, fullcomm, strata_comm, cylinder_comm):
        super().__init__(spbase_object, fullcomm, strata_comm, cylinder_comm)

    def make_windows(self):
        nscen = len(self.opt.all_scenario_names)
        if nscen == 0:
            raise RuntimeError(f"(rank: {self.cylinder_rank}), no local_scenarios")

        self.nscen = nscen
        vbuflen = 0
        self.nonant_per_scen = 0
        for s in self.opt.local_scenarios.values():
            vbuflen += len(s._mpisppy_data.nonant_indices)
        local_scen_count = len(self.opt.local_scenario_names)
        self.nonant_per_scen = int(vbuflen / local_scen_count)

        ## the _locals will also have the kill signal
        self.all_nonant_len = vbuflen
        self.all_eta_len = nscen*local_scen_count
        self._locals = np.zeros(nscen*local_scen_count + vbuflen + 1)
        self._coefs = np.zeros(nscen*(nscen + self.nonant_per_scen) + 1 + 1)
        self._new_locals = False

        # local, remote
        # send, receive
        self._make_windows(nscen*(self.nonant_per_scen + 1 + 1), nscen*local_scen_count + vbuflen)

    def _got_kill_signal(self):
        ''' returns True if a kill signal was received,
            and refreshes the array and _locals'''
        self._new_locals = self.spoke_from_hub(self._locals)
        kill = (self._locals[-1] == -1)
        return kill

    def prep_cs_cuts(self):
        # create a map scenario -> index, this index is used for various lists containing scenario dependent info.
        self.scenario_to_index = { scen : indx for indx, scen in enumerate(self.opt.all_scenario_names) }

        # create concrete model to use as pseudo-root
        self.opt.root = pyo.ConcreteModel()

        ##get the nonants off an arbitrary scenario
        arb_scen = self.opt.local_scenarios[self.opt.local_scenario_names[0]]
        non_ants = arb_scen._mpisppy_node_list[0].nonant_vardata_list

        # add copies of the nonanticipatory variables to the root problem
        # NOTE: the LShaped code expects the nonant vars to be in a particular
        #       order and with a particular *name*.
        #       We're also creating an index for reference against later 
        nonant_vid_to_copy_map = dict()
        root_vars = list()
        for v in non_ants:
            non_ant_copy = pyo.Var(name=v.name)
            self.opt.root.add_component(v.name, non_ant_copy)
            root_vars.append(non_ant_copy)
            nonant_vid_to_copy_map[id(v)] = non_ant_copy

        self.opt.root_vars = root_vars

        # create an index of these non_ant_copies to be in the same
        # order as PH, used below
        nonants = dict()
        for ndn_i, nonant in arb_scen._mpisppy_data.nonant_indices.items():
            vid = id(nonant)
            nonants[ndn_i] = nonant_vid_to_copy_map[vid]

        self.root_nonants = nonants
        self.opt.root.eta = pyo.Var(self.opt.all_scenario_names)

        self.opt.root.bender = LShapedCutGenerator()
        self.opt.root.bender.set_input(root_vars=self.opt.root_vars, 
                                            tol=1e-4, comm=self.cylinder_comm)
        self.opt.root.bender.set_ls(self.opt)

        ## the below for loop can take some time,
        ## so return early if we get a kill signal,
        ## but only after a barrier
        self.cylinder_comm.Barrier()
        if self.got_kill_signal():
            return

        # add the subproblems for all
        for scen in self.opt.all_scenario_names:
            subproblem_fn_kwargs = dict()

            # need to modify this to accept in user kwargs as well
            subproblem_fn_kwargs['scenario_name'] = scen
            self.opt.root.bender.add_subproblem(subproblem_fn=self.opt.create_subproblem,
                                                 subproblem_fn_kwargs=subproblem_fn_kwargs,
                                                 root_eta=self.opt.root.eta[scen],
                                                 subproblem_solver=self.opt.options["sp_solver"],
                                                 subproblem_solver_options=self.opt.options["sp_solver_options"])

        ## the above for loop can take some time,
        ## so return early if we get a kill signal,
        ## but only after a barrier
        self.cylinder_comm.Barrier()
        if self.got_kill_signal():
            return

        ## This call is blocking, depending on the
        ## configuration. This necessitates the barrier
        ## above.
        self.opt.set_eta_bounds()
        self._eta_lb_array = np.fromiter(
                (self.opt.valid_eta_lb[s] for s in self.opt.all_scenario_names),
                dtype='d', count=len(self.opt.all_scenario_names))
        self.make_eta_lb_cut()

    def make_eta_lb_cut(self):
        ## we'll be storing a matrix as an array
        ## row_len is the length of each row
        row_len = 1+1+len(self.root_nonants)
        all_coefs = np.zeros( self.nscen*row_len+1, dtype='d')
        for idx, k in enumerate(self.opt.all_scenario_names):
            ## cut_array -- [ constant, eta_coef, *nonant_coefs ]
            ## this cut  -- [ LB, -1, *0s ], i.e., -1*\eta + LB <= 0
            all_coefs[row_len*idx] = self._eta_lb_array[idx]
            all_coefs[row_len*idx+1] = -1
        self.spoke_to_hub(all_coefs)

    def make_cut(self):

        ## cache opt
        opt = self.opt

        ## unpack these the way they were packed:
        all_nonants_and_etas = self._locals
        nonants = dict()
        etas = dict()
        ci = 0
        for k, s in opt.local_scenarios.items():
            for ndn, i in s._mpisppy_data.nonant_indices:
                nonants[k, ndn, i] = all_nonants_and_etas[ci]
                ci += 1

        # get all the etas
        for k, s in opt.local_scenarios.items():
            for sn in opt.all_scenario_names:
                etas[k, sn] = all_nonants_and_etas[ci]
                ci += 1

        ## self.nscen == len(opt.all_scenario_names)
        # compute local min etas
        min_eta_vals = np.fromiter(( min(etas[k,sn] for k in opt.local_scenarios) \
                                       for sn in opt.all_scenario_names ),
                                    dtype='d', count=self.nscen)
        # Allreduce the etas to take the minimum
        global_eta_vals = np.empty(self.nscen, dtype='d')
        self.cylinder_comm.Allreduce(min_eta_vals, global_eta_vals, op=MPI.MIN)

        eta_lb_viol = (global_eta_vals + np.full_like(global_eta_vals, 1e-3) \
                        < self._eta_lb_array).any()
        if eta_lb_viol:
            self.make_eta_lb_cut()
            return

        # set the root etas to be the minimum from every scenario
        root_etas = opt.root.eta
        for idx, scen_name in enumerate(opt.all_scenario_names):
            root_etas[scen_name].set_value(global_eta_vals[idx])

        # sum the local nonants for average computation
        root_nonants = self.root_nonants

        local_nonant_sum = np.fromiter( ( sum(nonants[k, nname, ix] for k in opt.local_scenarios)
                                          for nname, ix in root_nonants),
                                          dtype='d', count=len(root_nonants) )


        # Allreduce the xhats to get averages
        global_nonant_sum = np.empty(len(local_nonant_sum), dtype='d')
        self.cylinder_comm.Allreduce(local_nonant_sum, global_nonant_sum, op = MPI.SUM)
        # need to divide through by the number of different spoke processes
        global_xbar = global_nonant_sum / self.nscen

        local_dist = np.array([0],dtype='d')
        local_winner = None
        # iterate through the ranks xhats to get the ranks maximum dist
        for i, k in enumerate(opt.local_scenarios):
            scenario_xhat = np.fromiter( (nonants[k, nname, ix] for nname, ix in root_nonants),
                                         dtype='d', count=len(root_nonants) )
            scenario_dist = np.linalg.norm(scenario_xhat - global_xbar)
            local_dist[0] = max(local_dist[0], scenario_dist)
            if local_winner is None:
                local_winner = k
            elif scenario_dist >= local_dist[0]:
                local_winner = k

        # Allreduce to find the biggest distance
        global_dist = np.empty(1, dtype='d')
        self.cylinder_comm.Allreduce(local_dist, global_dist, op=MPI.MAX)
        vote = np.array([-1], dtype='i')
        if local_dist[0] >= global_dist[0]:
            vote[0] = self.cylinder_comm.Get_rank()

        global_rank = np.empty(1, dtype='i')
        self.cylinder_comm.Allreduce(vote, global_rank, op=MPI.MAX)

        # if we are the winner, grab the xhat and bcast it to the other ranks
        if self.cylinder_comm.Get_rank() == global_rank[0]:
            farthest_xhat = np.fromiter( (nonants[local_winner, nname, ix] 
                                            for nname, ix in root_nonants),
                                         dtype='d', count=len(root_nonants) )
        else:
            farthest_xhat = np.zeros(len(root_nonants), dtype='d')

        self.cylinder_comm.Bcast(farthest_xhat, root=global_rank)

        # set the first stage in the lshape object to correspond to farthest_xhat
        for ci, k in enumerate(root_nonants):
            root_nonants[k].set_value(farthest_xhat[ci])

        # generate cuts
        cuts = opt.root.bender.generate_cut()

        # eta var_id map:
        eta_id_map = { id(var) : k for k,var in root_etas.items()}
        coef_dict = dict()
        feas_cuts = list()
        # package cuts, slightly silly in that we reconstruct the coefficients from the cuts
        # TODO: modify the lshaped_cuts method to have a separate generate_coeffs function
        for cut in cuts:
            repn = generate_standard_repn(cut.body)
            if len(repn.nonlinear_vars) > 0:
                raise RuntimeError("BendersCutGenerator returned nonlinear cut")

            ## create a map from id(var) to index in repn
            id_var_to_idx = { id(var) : i for i,var in enumerate(repn.linear_vars) }

            ## find the eta index
            for vid in eta_id_map:
                if vid in id_var_to_idx:
                    scen_name = eta_id_map[vid]
                    ## cut_array -- [ constant, eta_coef, *nonant_coefs ]
                    cut_array = [repn.constant, repn.linear_coefs[id_var_to_idx[vid]]]
                    # each eta_s should only appear at most once per set of cuts
                    del eta_id_map[vid]
                    # each variable should only appear once in repn.linear_vars
                    del id_var_to_idx[vid]
                    break
            else: # no break,
                # so no scenario recourse cost variables appear in the cut
                cut_array = [repn.constant, 0.]
                # we don't know what scenario,
                # but since eta_s is 0, it doesn't
                # matter
                scen_name = None

            ## be intentional about how these are loaded
            ## unloaded the same way
            ## root_vars is in the order PH expects
            ## (per above)
            for var in root_nonants.values():
                # each variable should only appear at most once in repn.linear_vars
                idx = id_var_to_idx.pop(id(var), None)
                if idx is not None:
                    cut_array.append(repn.linear_coefs[idx])
                else:
                    cut_array.append(0)

            if scen_name is not None:
                coef_dict[scen_name] = np.array(cut_array, dtype='d')
            else:
                feas_cuts.append( np.array(cut_array, dtype='d') )

        ## we'll be storing a matrix as an array
        ## row_len is the length of each row
        row_len = 1+1+len(root_nonants)
        all_coefs = np.zeros( self.nscen*row_len +1, dtype='d')
        for idx, k in enumerate(opt.all_scenario_names):
            if k in coef_dict:
                all_coefs[row_len*idx:row_len*(idx+1)] = coef_dict[k]
            elif feas_cuts:
                all_coefs[row_len*idx:row_len*(idx+1)] = feas_cuts.pop()
        self.spoke_to_hub(all_coefs)

    def main(self):
        # call main cut generation routine

        # prep cut generation
        self.prep_cs_cuts()

        # main loop
        while not (self.got_kill_signal()):
            if self._new_locals:
                self.make_cut()



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import pyomo.environ as pyo

# NOTE: a caller attaches the comms (e.g. pre_iter0)

import mpisppy.extensions.extension
from mpisppy.utils.sputils import _ScenTree


class XhatBase(mpisppy.extensions.extension.Extension):
    """
        Any inherited class must implement the preiter0, postiter etc. methods
        
        Args:
            opt (SPBase object): gives the problem that we bound

        Attributes:
          scenario_name_to_rank (dict of dict): nodes (i.e. comms) scen names
                keys are comms (i.e., tree nodes); values are dicts with keys
                that are scenario names and values that are ranks
    """
    def __init__(self, opt):
        super().__init__(opt)
        self.cylinder_rank = self.opt.cylinder_rank
        self.n_proc = self.opt.n_proc
        self.verbose = self.opt.options["verbose"]

        scen_count = len(opt.all_scenario_names)

        self.scenario_name_to_rank = opt.scenario_names_to_rank
        # dict: scenario names --> LOCAL rank number (needed mainly for xhat)
        
     #**********
    def _try_one(self, snamedict, solver_options=None, verbose=False, restore_nonants=True):
        """ try the scenario named sname in self.opt.local_scenarios
       Args:
            snamedict (dict): key: scenario tree non-leaf name, val: scen name
                            the scen name can be None if it is not local
            solver_options (dict): passed through to the solver
            verbose (boolean): controls output
            restore_nonants (bool): if True, restores the nonants to their original
                                    values in all scenarios. If False, leaves the
                                    nonants as they are in the tried scenario
       NOTE: The solve loop is with fixed nonants so W and rho do not
             matter to the optimization. When we want to check the obj
             value we need to drop them, but we don't need to re-optimize
             since all other Vars are optimized already with the nonants
             fixed.
        Returns:
             obj (float or None): the expected value for sname as xhat or None
        """
        xhats = dict()  # to pass to _fix_nonants
        self.opt._save_nonants()  # (BTW: for all local scenarios)
        # For now, we are going to treat two-stage as a special case
        if len(snamedict) == 1:
            sname = snamedict["ROOT"]  # also serves as an assert
            if sname in self.opt.local_scenarios:
                xhat = self.opt.local_scenarios[sname]._mpisppy_data.nonant_cache
            else:
                xhat = None
            src_rank = self.scenario_name_to_rank["ROOT"][sname]
            try:
                xhats["ROOT"] = self.comms["ROOT"].bcast(xhat, root=src_rank)
            except:
                print("rank=",self.cylinder_rank, "xhats bcast failed on src_rank={}"\
                      .format(src_rank))
                print("root comm size={}".format(self.comms["ROOT"].size))
                raise
        else:  # multi-stage
            # assemble parts and put it in xhats
            # send to ranks in the comm or receive ANY_SOURCE
            # (for closest do allreduce with loc operator) rank
            nlens = dict()
            cistart = dict()  # ci start for each local node
            for k, s in self.opt.local_scenarios.items():
                for nnode in s._mpisppy_node_list:
                    ndn = nnode.name
                    if ndn not in cistart:
                        # NOTE: _mpisppy_data.cistart is defined in SPBase._attach_nlens()
                        #       and only used here
                        cistart[ndn] = s._mpisppy_data.cistart[ndn]
                    if ndn not in nlens:
                        nlens[ndn] = s._mpisppy_data.nlens[ndn]
                    if ndn not in xhats:
                        xhats[ndn] = None
                    if ndn not in snamedict:
                        raise RuntimeError(f"{ndn} not in snamedict={snamedict}")
                    if snamedict[ndn] == k:
                        # cache lists are just concated node lists
                        xhats[ndn] = [s._mpisppy_data.nonant_cache[i+cistart[ndn]]
                                      for i in range(nlens[ndn])]
            for ndn in cistart:  # local nodes
                if snamedict[ndn] not in self.scenario_name_to_rank[ndn]:
                    print (f"For ndn={ndn}, snamedict[ndn] not in "
                           "self.scenario_name_to_rank[ndn]")
                    print(f"snamedict[ndn]={snamedict[ndn]}")
                    print(f"self.scenario_name_to_rank[ndn]={self.scenario_name_to_rank[ndn]}")
                    raise RuntimeError("Bad scenario selection for xhat")
                src_rank = self.scenario_name_to_rank[ndn][snamedict[ndn]]
                try:
                    xhats[ndn] = self.comms[ndn].bcast(xhats[ndn], root=src_rank)
                except:
                    print("rank=",self.cylinder_rank, "xhats bcast failed on ndn={}, src_rank={}"\
                          .format(ndn,src_rank))
                    raise
    
            # assemble xhat (which is a nonants dict) from xhats
            for ndn in xhats:
                for i in range(cistart[ndn], nlens[ndn]):
                    xhats[ndn] = xhats[ndn]
        # The save is done above
        self.opt._fix_nonants(xhats)  # (BTW: for all local scenarios)

        # Special Tee option for xhat
        sopt = solver_options
        Tee=False
        if solver_options is not None and "Tee" in solver_options:
            sopt = dict(solver_options)
            Tee = sopt["Tee"]
            del sopt["Tee"]

        # NOTE: for APH we may need disable_pyomo_signal_handling
        self.opt.solve_loop(solver_options=sopt,
                           #dis_W=True, dis_prox=True,
                           verbose=verbose,
                           tee=Tee)

        infeasP = self.opt.infeas_prob()
        if infeasP != 0.:
            # restoring does no harm
            # if this solution is infeasible
            self.opt._restore_nonants()
            return None
        else:
            if verbose and src_rank == self.cylinder_rank:
                print("   Feasible xhat found:")
                self.opt.local_scenarios[sname].pprint()
            obj = self.opt.Eobjective(verbose=verbose)
            if restore_nonants:
                self.opt._restore_nonants()
            return obj

    #**********
    def csv_nonants(self, snamedict, fname):
        """ write the non-ants in csv format to files based on the file name
            (we will over-write files if they already exists)
            Args:
               snamedic (str): the names of the scenarios to use
               fname (str): the full name of the file to which to write
        """
        # only the rank with the requested scenario writes
        for ndn, sname in snamedict.items():
            if sname not in self.opt.local_scenarios:
                continue
            scen = self.opt.local_scenarios[sname]
            with open(fname+"_"+ndn+"_"+sname+".csv", "w") as f:
                for node in scen._mpisppy_node_list:
                    if node.name == ndn:
                        break
                nlens = scen._mpisppy_data.nlens
                f.write(ndn)
                for i in range(nlens[ndn]):
                    vardata = node.nonant_vardata_list[i]
                    f.write(', "'+vardata.name+'", '+str(vardata._value))
                f.write("\n")

    #**********
    def csv_allvars(self, snamedict, fname):
        """ write all Vars in csv format to files based on the file name
            (we will over-write files if they already exists)
            Args:
               snamedict (dict): scenario names
               fname (str): the full name of the file to which to write
        """
        # only the rank with the requested scenario writes
        for ndn, sname in snamedict.items():
            if sname not in self.opt.local_scenarios:
                continue
            scen = self.opt.local_scenarios[sname]
            for node in scen._mpisppy_node_list:
                if node.name == ndn:
                    break
                with open(fname+"_"+ndn+"_"+sname,"w") as f:
                    for ((v_name, v_index), v_data)\
                        in scen.component_data_iterindex(pyo.Var, active=True):
                        f.write(v_name + ", " + str(pyo.value(v_data)) + "\n")

    """ Functions to be called by xhat extensions. This is just code
    common to all, or almost all xhat extensions. It happens to be the
    case that whatever extobject that gets passed in will have been
    derived from XhatBase, but that is not why the code is here. It was
    factored simply for the usual reasons to factor code.  """

    def xhat_common_post_everything(self, extname, obj, snamedict, restored_nonants):
        """ Code that most xhat post_everything routines will want to call.
        Args:
            extname (str): the name of the extension for reporting
            obj (float): the xhat objective function
            snamedict (dict): the (scenario) names upon which xhat is based
            restored_nonants (bool): if the restore_nonants flag was True on the last
                call to _try_one.
        """
        if (obj is not None) and (not restored_nonants):
            # a tree solution is available
            self.opt.tree_solution_available = True
            self.opt.first_stage_solution_available = True
        if (obj is not None) and (self.opt.spcomm is not None):
            self.opt.spcomm.BestInnerBound = self.opt.spcomm.InnerBoundUpdate(obj, char='E')
        if self.cylinder_rank == 0 and self.verbose:
            print ("****", extname ,"Used scenarios",
                   str(snamedict),"to get xhat Eobj=",obj)

        if "csvname" in self.options:
            self.csv_nonants(snamedict, self.options["csvname"])

        if "dump_prefix" in self.options:
            prefpref = self.options["dump_prefix"]
            pref = extname
            self.csv_nonants(snamedict, prefpref + "_nonant_" + pref)
            self.csv_allvars(snamedict, prefpref + "_allvars_" + pref)

    def post_iter0(self):
        # the base class needs this
        self.comms = self.opt.comms
        
"""
May 2020, DLW
Simple tree with branching factors (arbitrary trees later)

==============

Background:

0. We have a scenario tree defined by a ROOT and branching factors for
each stage beyond the first stage. E.g., for a three stage tree the
branching factors might be [2,3].We refer to the final stage tree
nodes and "leaf" nodes and all other nodes as "non-leaf" nodes. In the
example, there are three non-leaf nodes: ROOT, plus the two nodes at
in the second stage.

1. One path through the tree defines a scenario, so the number of
scenarios is the product of the branching factors.

2. We want to allow an arbitrary number of stages but, realistically, there would be at most 6.

3. Processing is done on ranks and scenarios must be assigned to ranks (MPI terminology).

4. We want as much flexibility as possible when determining the number of ranks, but
it might not be unlimited flexibility.

5. For a given rank, the scenarios assigned to it must all be in the same non-leaf nodes. The
reasons have to do with how we will use MPI comms: There will be a comm for each non-leaf node
and reductions will be done for the node on its corresponding comm. This is critical.

=================

Input: number of ranks, list of branching factors, list of scenario names

output: An error message concerning the relationship between the inputs
or

A dictionary mapping each non-leaf node name in the scenario tree
to a dictionary mapping scenario name to rank number.

(There is redundancy in the inputs and outputs, but both are handy for use.)

=================

Naming conventions:

= non-leaf nodes other than ROOT are named <parent>_nodenum so for the
example, there are three non-leaf nodes named ROOT, ROOT_1, and ROOT_2

= we happen to be given a list of scenario names that has the same
length as the number of scenarios (i.e., the product of the branching
factors) to recover (or construct) the scenario names, the tree is
traversed depth-first.  As a practical matter, they will have names
like Scenx or Scenariox where x is replaced by a serial number.

==================

Notes:

The outer-most non-leaf nodes completely determine the assignment of scenarios
to ranks.

So maybe we should flip it around and assign the ranks to the outer-most
non-leaf tree nodes.

A rank cannot span two such nodes, but a node can have two ranks.
"""



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
from math import inf, isfinite
from pyomo.core.expr.numeric_expr import LinearExpression
from mpisppy.cylinders.hub import PHHub
from mpisppy.cylinders.cross_scen_spoke import CrossScenarioCutSpoke

import numpy as np
import mpisppy.utils.sputils as sputils

class CrossScenarioHub(PHHub):
    def setup_hub(self):
        super().setup_hub()
        if self.opt.multistage:
            raise RuntimeError('CrossScenarioHub only supports '
                                'two-stage models at this time')
        idx = self.cut_gen_spoke_index
        self.all_nonants_and_etas = np.zeros(self.local_lengths[idx - 1] + 1)

        arb_scen = self.opt.local_scenarios[self.opt.local_scenario_names[0]]
        # get an arbitrary scenario for the 
        # number of nonant variables
        self.nonant_len = len(arb_scen._mpisppy_data.nonant_indices)

        # save the best bounds so far
        self.best_inner_bound = inf
        self.best_outer_bound = -inf

        # helping the extension track cuts
        self.new_cuts = False

    def initialize_spoke_indices(self):
        super().initialize_spoke_indices()
        for (i, spoke) in enumerate(self.spokes):
            if spoke["spoke_class"] == CrossScenarioCutSpoke:
                self.cut_gen_spoke_index = i + 1

    def sync(self):
        super().sync()
        self.send_to_cross_cuts()
        self.get_from_cross_cuts()

    def get_from_cross_cuts(self):
        idx = self.cut_gen_spoke_index
        receive_buffer = np.empty(self.remote_lengths[idx - 1] + 1, dtype="d") # Must be doubles
        is_new = self.hub_from_spoke(receive_buffer, idx)
        if is_new:
            self.make_cuts(receive_buffer)

    def send_to_cross_cuts(self):
        idx = self.cut_gen_spoke_index

        # get the stuff we want to send
        self.opt._save_nonants()
        ci = 0  ## index to self.nonant_send_buffer

        # get all the nonants
        all_nonants_and_etas = self.all_nonants_and_etas
        for k, s in self.opt.local_scenarios.items():
            for xvar in s._mpisppy_data.nonant_indices.values():
                all_nonants_and_etas[ci] = xvar._value
                ci += 1

        # get all the etas
        for k, s in self.opt.local_scenarios.items():
            for sn in self.opt.all_scenario_names:
                all_nonants_and_etas[ci] = s._mpisppy_model.eta[sn]._value
                ci += 1
        self.hub_to_spoke(all_nonants_and_etas, idx)


    def make_cuts(self, coefs):
        # take the coefficient array and assemble cuts accordingly

        # this should have already been set in the extension !
        opt = self.opt

        # rows are 
        # [ const, eta_coeff, *nonant_coeffs ]
        row_len = 1+1+self.nonant_len
        outer_iter = int(coefs[-1])

        bundling = opt.bundling
        if opt.bundling:
            for bn,b in opt.local_subproblems.items():
                persistent_solver = sputils.is_persistent(b._solver_plugin)
                ## get an arbitrary scenario
                s = opt.local_scenarios[b.scen_list[0]]
                for idx, k in enumerate(opt.all_scenario_names):
                    row = coefs[row_len*idx:row_len*(idx+1)]
                    # the row could be all zeros,
                    # which doesn't do anything
                    if (row == 0.).all():
                        continue
                    # rows are 
                    # [ const, eta_coeff, *nonant_coeffs ]
                    linear_const = row[0]
                    linear_coefs = list(row[1:])
                    linear_vars = [b._mpisppy_model.eta[k]]

                    for ndn_i in s._mpisppy_data.nonant_indices:
                        ## for bundles, we add the constrains only
                        ## to the reference first stage variables
                        linear_vars.append(b.ref_vars[ndn_i])

                    cut_expr = LinearExpression(constant=linear_const, linear_coefs=linear_coefs,
                                                linear_vars=linear_vars)
                    b._mpisppy_model.benders_cuts[outer_iter, k] = (None, cut_expr, 0)
                    if persistent_solver:
                        b._solver_plugin.add_constraint(b._mpisppy_model.benders_cuts[outer_iter, k])

        else:
            for sn,s in opt.local_subproblems.items():
                persistent_solver = sputils.is_persistent(s._solver_plugin)
                for idx, k in enumerate(opt.all_scenario_names):
                    row = coefs[row_len*idx:row_len*(idx+1)]
                    # the row could be all zeros,
                    # which doesn't do anything
                    if (row == 0.).all():
                        continue
                    # rows are 
                    # [ const, eta_coeff, *nonant_coeffs ]
                    linear_const = row[0]
                    linear_coefs = list(row[1:])
                    linear_vars = [s._mpisppy_model.eta[k]]
                    linear_vars.extend(s._mpisppy_data.nonant_indices.values())

                    cut_expr = LinearExpression(constant=linear_const, linear_coefs=linear_coefs,
                                                linear_vars=linear_vars)
                    s._mpisppy_model.benders_cuts[outer_iter, k] = (None, cut_expr, 0.)
                    if persistent_solver:
                        s._solver_plugin.add_constraint(s._mpisppy_model.benders_cuts[outer_iter, k])

        # NOTE: the LShaped code negates the objective, so
        #       we do the same here for consistency
        ib = self.BestInnerBound
        ob = self.BestOuterBound
        if not opt.is_minimizing:
            ib = -ib
            ob = -ob
        add_cut = (isfinite(ib) or isfinite(ob)) and \
                ((ib < self.best_inner_bound) or (ob > self.best_outer_bound))
        if add_cut:
            self.best_inner_bound = ib
            self.best_outer_bound = ob
            for sn,s in opt.local_subproblems.items():
                persistent_solver = sputils.is_persistent(s._solver_plugin)
                prior_outer_iter = list(s._mpisppy_model.inner_bound_constr.keys())
                s._mpisppy_model.inner_bound_constr[outer_iter] = (ob, s._mpisppy_model.EF_obj, ib)
                if persistent_solver:
                    s._solver_plugin.add_constraint(s._mpisppy_model.inner_bound_constr[outer_iter])
                # remove other ib constraints (we only need the tightest)
                for it in prior_outer_iter:
                    if persistent_solver:
                        s._solver_plugin.remove_constraint(s._mpisppy_model.inner_bound_constr[it])
                    del s._mpisppy_model.inner_bound_constr[it]

        ## helping the extention track cuts
        self.new_cuts = True



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
''' Implementation of the Frank-Wolfe Progressive Hedging (FW-PH) algorithm
    described in the paper:

    N. Boland et al., "Combining Progressive Hedging with a Frank-Wolfe method
    to compute Lagrangian dual bounds in stochastic mixed-integer programming".
    SIAM J. Optim. 28(2):1312--1336, 2018.

    Current implementation supports parallelism and bundling.

    Does not support:
         1. The use of extensions
         2. The solution of models with more than two stages
         3. Simultaneous use of bundling and user-specified initial points

    The FWPH algorithm allows the user to specify an initial set of points
    (notated V^0_s in Boland) to approximate the convex hull of each scenario
    subproblem. Points are specified via a callable which takes as input a
    scenario name (str) and outputs a list of dicts. Each dict corresponds to a
    point, and is formatted

        point_dict[variable_name] = variable_value

    (e.g. point_dict['Allocation[2,5]'] = 1.7). The order of this list
    matters--the first point in the list is used to compute the initial dual
    weights and xBar value (in Boland notation, the first point in the list is
    the point (x0_s,y0_s) in Algorithm 3). The callable is passed to FWPH as an
    option in the FW_options dict, with the key "point_creator". The
    point_creator callable may also take an optional argument,
    "point_creator_data", which may be any data type, and contain any
    information needed to create the initial point set. The point_creator_data
    is passed to FWPH as an option in the FW_options dict, with the key
    "point_creator_data".

    See fwph_sslp.py for an example of user-specified point creation.
'''

import mpisppy.phbase
import mpisppy.utils.sputils as sputils
import numpy as np
import pyomo.environ as pyo
import time
import re # For manipulating scenario names

from mpi4py import MPI
from pyomo.repn.standard_repn import generate_standard_repn
from mpisppy.utils.sputils import find_active_objective
from pyomo.core.expr.visitor import replace_expressions
from pyomo.core.expr.numeric_expr import LinearExpression

class FWPH(mpisppy.phbase.PHBase):
    
    def __init__(
        self,
        PH_options,
        FW_options,
        all_scenario_names,
        scenario_creator,
        scenario_denouement=None,
        all_nodenames=None,
        mpicomm=None,
        scenario_creator_kwargs=None,
        PH_converger=None,
        rho_setter=None,
    ):
        super().__init__(
            PH_options, 
            all_scenario_names,
            scenario_creator,
            scenario_denouement,
            all_nodenames=all_nodenames,
            mpicomm=mpicomm,
            scenario_creator_kwargs=scenario_creator_kwargs,
            extensions=None,
            extension_kwargs=None,
            PH_converger=PH_converger,
            rho_setter=rho_setter,
        )

        self._init(FW_options)

    def _init(self, FW_options):
        self.FW_options = FW_options
        self._options_checks_fw()
        self.vb = True
        if ('FW_verbose' in self.FW_options):
            self.vb = self.FW_options['FW_verbose']

    def fw_prep(self):
        self.PH_Prep(attach_duals=True, attach_prox=False)
        self._check_for_multistage()
        self.subproblem_creation(self.options['verbose'])
        self._output_header()

        if ('point_creator' in self.FW_options):
            # The user cannot both specify and point_creator and use bundles.
            # At this point, we have already checked for that possibility, so
            # we can safely use the point_creator without further checks for
            # bundles.
            self._create_initial_points()
            check = True if 'check_initial_points' not in self.FW_options \
                         else self.FW_options['check_initial_points']
            if (check):
                self._check_initial_points()
            self._create_solvers()
            self._use_rho_setter(verbose and self.cylinder_rank==0)
            self._initialize_MIP_var_values()
            best_bound = -np.inf if self.is_minimizing else np.inf
        else:
            trivial_bound = self.Iter0()
            secs = time.time() - self.t0
            self._output(0, trivial_bound, trivial_bound, np.nan, secs)
            best_bound = trivial_bound

        if ('mip_solver_options' in self.FW_options):
            self._set_MIP_solver_options()

        # Lines 2 and 3 of Algorithm 3 in Boland
        self.Compute_Xbar(self.options['verbose'])
        self.Update_W(self.options['verbose'])

        # Necessary pre-processing steps
        # We disable_W so they don't appear
        # in the MIP objective when _set_QP_objective
        # snarfs it for the QP
        self._disable_W()
        self._attach_MIP_vars()
        self._initialize_QP_subproblems()
        self._attach_indices()
        self._attach_MIP_QP_maps()
        self._set_QP_objective()
        self._initialize_QP_var_values()
        self._swap_nonant_vars()
        self._reenable_W()

        if (self.PH_converger):
            self.convobject = self.PH_converger(self, self.cylinder_rank, self.n_proc)

        return best_bound

    def fwph_main(self):
        self.t0 = time.time()
        best_bound = self.fw_prep()

        # FWPH takes some time to initialize
        # If run as a spoke, check for convergence here
        if self.spcomm and self.spcomm.is_converged():
            return None, None, None

        # The body of the algorithm
        for itr in range(self.options['PHIterLimit']):
            self._PHIter = itr
            self._local_bound = 0
            for name in self.local_subproblems:
                dual_bound = self.SDM(name)
                self._local_bound += self.local_subproblems[name]._mpisppy_probability * \
                                     dual_bound
            self._compute_dual_bound()
            if (self.is_minimizing):
                best_bound = np.maximum(best_bound, self._local_bound)
            else:
                best_bound = np.minimum(best_bound, self._local_bound)

            ## Hubs/spokes take precedence over convergers
            if self.spcomm:
                if self.spcomm.is_converged():
                    secs = time.time() - self.t0
                    self._output(itr+1, self._local_bound, 
                                 best_bound, np.nan, secs)
                    if (self.cylinder_rank == 0 and self.vb):
                        print('FWPH converged to user-specified criteria')
                    break
                self.spcomm.sync()
            if (self.PH_converger):
                self.Compute_Xbar(self.options['verbose'])
                diff = self.convobject.convergence_value()
                if (self.convobject.is_converged()):
                    secs = time.time() - self.t0
                    self._output(itr+1, self._local_bound, 
                                 best_bound, diff, secs)
                    if (self.cylinder_rank == 0 and self.vb):
                        print('FWPH converged to user-specified criteria')
                    break
            else: # Convergence check from Boland
                diff = self._conv_diff()
                self.Compute_Xbar(self.options['verbose'])
                if (diff < self.options['convthresh']):
                    secs = time.time() - self.t0
                    self._output(itr+1, self._local_bound, 
                                 best_bound, diff, secs)
                    if (self.cylinder_rank == 0 and self.vb):
                        print('PH converged based on standard criteria')
                    break

            secs = time.time() - self.t0
            self._output(itr+1, self._local_bound, best_bound, diff, secs)
            self.Update_W(self.options['verbose'])
            timed_out = self._is_timed_out()
            if (self._is_timed_out()):
                if (self.cylinder_rank == 0 and self.vb):
                    print('Timeout.')
                break

        self._swap_nonant_vars_back()
        weight_dict = self._gather_weight_dict() # None if rank != 0
        xbars_dict  = self._get_xbars() # None if rank != 0
        return itr+1, weight_dict, xbars_dict

    def SDM(self, model_name):
        '''  Algorithm 2 in Boland et al. (with small tweaks)
        '''
        mip = self.local_subproblems[model_name]
        qp  = self.local_QP_subproblems[model_name]
    
        # Set the QP dual weights to the correct values If we are bundling, we
        # initialize the QP dual weights to be the dual weights associated with
        # the first scenario in the bundle (this should be okay, because each
        # scenario in the bundle has the same dual weights, analytically--maybe
        # a numerical problem).
        arb_scen_mip = self.local_scenarios[mip.scen_list[0]] \
                       if self.bundling else mip
        for (node_name, ix) in arb_scen_mip._mpisppy_data.nonant_indices:
            qp._mpisppy_model.W[node_name, ix]._value = \
                arb_scen_mip._mpisppy_model.W[node_name, ix].value

        alpha = self.FW_options['FW_weight']
        # Algorithm 3 line 6
        xt = {ndn_i:
            (1 - alpha) * pyo.value(arb_scen_mip._mpisppy_model.xbars[ndn_i])
            + alpha * pyo.value(xvar)
            for ndn_i, xvar in arb_scen_mip._mpisppy_data.nonant_indices.items()
            }

        for itr in range(self.FW_options['FW_iter_limit']):
            # Algorithm 2 line 4
            mip_source = mip.scen_list if self.bundling else [model_name]
            for scenario_name in mip_source:
                scen_mip = self.local_scenarios[scenario_name]
                for ndn_i, nonant in scen_mip._mpisppy_data.nonant_indices.items():
                    x_source = xt[ndn_i] if itr==0 \
                               else nonant._value
                    scen_mip._mpisppy_model.W[ndn_i]._value = (
                        qp._mpisppy_model.W[ndn_i]._value
                        + scen_mip._mpisppy_model.rho[ndn_i]._value
                        * (x_source
                        -  scen_mip._mpisppy_model.xbars[ndn_i]._value))

            # Algorithm 2 line 5
            if (sputils.is_persistent(mip._solver_plugin)):
                mip_obj = find_active_objective(mip)
                mip._solver_plugin.set_objective(mip_obj)
            mip_results = mip._solver_plugin.solve(mip)
            self._check_solve(mip_results, model_name + ' (MIP)')

            # Algorithm 2 lines 6--8
            obj = find_active_objective(mip)
            if (itr == 0):
                if (self.is_minimizing):
                    dual_bound = mip_results.Problem[0].Lower_bound
                else:
                    dual_bound = mip_results.Problem[0].Upper_bound

            # Algorithm 2 line 9 (compute \Gamma^t)
            val0 = pyo.value(obj)
            new  = replace_expressions(obj.expr, mip._mpisppy_data.mip_to_qp)
            val1 = pyo.value(new)
            obj.expr = replace_expressions(new, qp._mpisppy_data.qp_to_mip)
            if abs(val0) > 1e-9:
                stop_check = (val1 - val0) / abs(val0) # \Gamma^t in Boland, but normalized
            else:
                stop_check = val1 - val0 # \Gamma^t in Boland
            stop_check_tol = self.FW_options["stop_check_tol"]\
                             if "stop_check_tol" in self.FW_options else 1e-4
            if (self.is_minimizing and stop_check < -stop_check_tol):
                print('Warning (fwph): convergence quantity Gamma^t = '
                     '{sc:.2e} (should be non-negative)'.format(sc=stop_check))
                print('Try decreasing the MIP gap tolerance and re-solving')
            elif (not self.is_minimizing and stop_check > stop_check_tol):
                print('Warning (fwph): convergence quantity Gamma^t = '
                     '{sc:.2e} (should be non-positive)'.format(sc=stop_check))
                print('Try decreasing the MIP gap tolerance and re-solving')

            self._add_QP_column(model_name)
            if (sputils.is_persistent(qp._QP_solver_plugin)):
                qp_obj = find_active_objective(qp)
                qp._QP_solver_plugin.set_objective(qp_obj)
            qp_results = qp._QP_solver_plugin.solve(qp)
            self._check_solve(qp_results, model_name + ' (QP)')

            if (stop_check < self.FW_options['FW_conv_thresh']):
                break

        # Re-set the mip._mpisppy_model.W so that the QP objective 
        # is correct in the next major iteration
        mip_source = mip.scen_list if self.bundling else [model_name]
        for scenario_name in mip_source:
            scen_mip = self.local_scenarios[scenario_name]
            for (node_name, ix) in scen_mip._mpisppy_data.nonant_indices:
                scen_mip._mpisppy_model.W[node_name, ix]._value = \
                    qp._mpisppy_model.W[node_name, ix]._value

        return dual_bound

    def _add_QP_column(self, model_name):
        ''' Add a column to the QP, with values taken from the most recent MIP
            solve.
        '''
        mip = self.local_subproblems[model_name]
        qp  = self.local_QP_subproblems[model_name]
        solver = qp._QP_solver_plugin
        persistent = sputils.is_persistent(solver)

        if hasattr(solver, 'add_column'):
            new_var = qp.a.add()
            coef_list = [1.]
            constr_list = [qp.sum_one]
            target = mip.ref_vars if self.bundling else mip.nonant_vars
            for (node, ix) in qp.eqx.index_set():
                coef_list.append(target[node, ix].value)
                constr_list.append(qp.eqx[node, ix])
            for key in mip._mpisppy_model.y_indices:
                coef_list.append(mip.leaf_vars[key].value)
                constr_list.append(qp.eqy[key])
            solver.add_column(qp, new_var, 0, constr_list, coef_list)
            return

        # Add new variable and update \sum a_i = 1 constraint
        new_var = qp.a.add() # Add the new convex comb. variable
        if (persistent):
            solver.add_var(new_var)
            solver.remove_constraint(qp.sum_one)
            qp.sum_one._body += new_var
            solver.add_constraint(qp.sum_one)
        else:
            qp.sum_one._body += new_var

        target = mip.ref_vars if self.bundling else mip.nonant_vars
        for (node, ix) in qp.eqx.index_set():
            if (persistent):
                solver.remove_constraint(qp.eqx[node, ix])
                qp.eqx[node, ix]._body += new_var * target[node, ix].value
                solver.add_constraint(qp.eqx[node, ix])
            else:
                qp.eqx[node, ix]._body += new_var * target[node,ix].value
        for key in mip._mpisppy_model.y_indices:
            if (persistent):
                solver.remove_constraint(qp.eqy[key])
                qp.eqy[key]._body += new_var * pyo.value(mip.leaf_vars[key])
                solver.add_constraint(qp.eqy[key])
            else:
                qp.eqy[key]._body += new_var * pyo.value(mip.leaf_vars[key])

    def _attach_indices(self):
        ''' Attach the fields x_indices and y_indices to the model objects in
            self.local_subproblems (not self.local_scenarios, nor
            self.local_QP_subproblems).

            x_indices is a list of tuples of the form...
                (scenario name, node name, variable index) <-- bundling
                (node name, variable index)                <-- no bundling
            y_indices is a list of tuples of the form...
                (scenario_name, "LEAF", variable index)    <-- bundling
                ("LEAF", variable index)                   <-- no bundling
            
            Must be called after the subproblems (MIPs AND QPs) are created.
        '''
        for name in self.local_subproblems.keys():
            mip = self.local_subproblems[name]
            qp  = self.local_QP_subproblems[name]
            if (self.bundling):
                x_indices = [(scenario_name, node_name, ix)
                    for scenario_name in mip.scen_list
                    for (node_name, ix) in 
                        self.local_scenarios[scenario_name]._mpisppy_data.nonant_indices]
                y_indices = [(scenario_name, 'LEAF', ix)
                    for scenario_name in mip.scen_list
                    for ix in range(mip.num_leaf_vars[scenario_name])]
            else:
                x_indices = mip._mpisppy_data.nonant_indices.keys()
                y_indices = [('LEAF', ix) for ix in range(len(qp.y))]

            y_indices = pyo.Set(initialize=y_indices)
            y_indices.construct()
            x_indices = pyo.Set(initialize=x_indices)
            x_indices.construct()
            mip._mpisppy_model.x_indices = x_indices
            mip._mpisppy_model.y_indices = y_indices

    def _attach_MIP_QP_maps(self):
        ''' Create dictionaries that map MIP variable ids to their QP
            counterparts, and vice versa.
        '''
        for name in self.local_subproblems.keys():
            mip = self.local_subproblems[name]
            qp  = self.local_QP_subproblems[name]

            mip._mpisppy_data.mip_to_qp = {id(mip.nonant_vars[key]): qp.x[key]
                                for key in mip._mpisppy_model.x_indices}
            mip._mpisppy_data.mip_to_qp.update({id(mip.leaf_vars[key]): qp.y[key]
                                for key in mip._mpisppy_model.y_indices})
            qp._mpisppy_data.qp_to_mip = {id(qp.x[key]): mip.nonant_vars[key]
                                for key in mip._mpisppy_model.x_indices}
            qp._mpisppy_data.qp_to_mip.update({id(qp.y[key]): mip.leaf_vars[key]
                                for key in mip._mpisppy_model.y_indices})

    def _attach_MIP_vars(self):
        ''' Create a list indexed (node_name, ix) for all the MIP
            non-anticipative and leaf variables, so that they can be easily
            accessed when adding columns to the QP.
        '''
        if (self.bundling):
            for (bundle_name, EF) in self.local_subproblems.items():
                EF.nonant_vars = dict()
                EF.leaf_vars   = dict()
                EF.num_leaf_vars = dict() # Keys are scenario names
                for scenario_name in EF.scen_list:
                    mip = self.local_scenarios[scenario_name]
                    # Non-anticipative variables
                    nonant_dict = {(scenario_name, ndn, ix): nonant
                        for (ndn,ix), nonant in mip._mpisppy_data.nonant_indices.items()}
                    EF.nonant_vars.update(nonant_dict)
                    # Leaf variables
                    leaf_vars = self._get_leaf_vars(mip)
                    leaf_var_dict = {(scenario_name, 'LEAF', ix): 
                        leaf_vars[ix] for ix in range(len(leaf_vars))}
                    EF.leaf_vars.update(leaf_var_dict)
                    EF.num_leaf_vars[scenario_name] = len(leaf_vars)
                    # Reference variables are already attached: EF.ref_vars
                    # indexed by (node_name, index)
        else:
            for (name, mip) in self.local_scenarios.items():
                mip.nonant_vars = mip._mpisppy_data.nonant_indices
                leaf_vars = self._get_leaf_vars(mip)
                mip.leaf_vars = { ('LEAF', ix): 
                    leaf_vars[ix] for ix in range(len(leaf_vars))
                }

    def _check_for_multistage(self):
        if self.multistage:
            raise RuntimeError('The FWPH algorithm only supports '
                               'two-stage models at this time.')

    def _check_initial_points(self):
        ''' If t_max (i.e. the inner iteration limit) is set to 1, then the
            initial point set must satisfy the additional condition (17) in
            Boland et al. This function verifies that condition (17) is
            satisfied by solving a linear program (similar to the Phase I
            auxiliary LP in two-phase simplex).

            This function is only called by a single rank, which MUST be rank
            0. The rank 0 check happens before this function is called.
        '''
        # Need to get the first-stage variable names (as supplied by the user)
        # by picking them off of any random scenario that's laying around.
        arb_scenario = list(self.local_scenarios.keys())[0]
        arb_mip = self.local_scenarios[arb_scenario]
        root = arb_mip._mpisppy_node_list[0]
        stage_one_var_names = [var.name for var in root.nonant_vardata_list]

        init_pts = self.comms['ROOT'].gather(self.local_initial_points, root=0)
        if (self.cylinder_rank != 0):
            return

        print('Checking initial points...', end='', flush=True)
        
        points = {key: value for block in init_pts 
                             for (key, value) in block.items()}
        scenario_names = points.keys()
        num_scenarios = len(points)

        # Some index sets we will need..
        conv_ix = [(scenario_name, var_name) 
                    for scenario_name in scenario_names
                    for var_name in stage_one_var_names]
        conv_coeff = [(scenario_name, ix) 
                    for scenario_name in scenario_names
                    for ix in range(len(points[scenario_name]))]

        aux = pyo.ConcreteModel()
        aux.x = pyo.Var(stage_one_var_names, within=pyo.Reals)
        aux.slack_plus = pyo.Var(conv_ix, within=pyo.NonNegativeReals)
        aux.slack_minus = pyo.Var(conv_ix, within=pyo.NonNegativeReals)
        aux.conv = pyo.Var(conv_coeff, within=pyo.NonNegativeReals)

        def sum_one_rule(model, scenario_name):
            return pyo.quicksum(model.conv[scenario_name,ix] \
                for ix in range(len(points[scenario_name]))) == 1
        aux.sum_one = pyo.Constraint(scenario_names, rule=sum_one_rule)

        def conv_rule(model, scenario_name, var_name):
            return model.x[var_name] \
                + model.slack_plus[scenario_name, var_name] \
                - model.slack_minus[scenario_name, var_name] \
                == pyo.quicksum(model.conv[scenario_name, ix] *
                    points[scenario_name][ix][var_name] 
                    for ix in range(len(points[scenario_name])))
        aux.comb = pyo.Constraint(conv_ix, rule=conv_rule)

        obj_expr = pyo.quicksum(aux.slack_plus.values()) \
                   + pyo.quicksum(aux.slack_minus.values())
        aux.obj = pyo.Objective(expr=obj_expr, sense=pyo.minimize)

        solver = pyo.SolverFactory(self.FW_options['solvername'])
        results = solver.solve(aux)
        self._check_solve(results, 'Auxiliary LP')

        check_tol = self.FW_options['check_tol'] \
                        if 'check_tol' in self.FW_options.keys() else 1e-4
        if (pyo.value(obj_expr) > check_tol):
            print('error.')
            raise ValueError('The specified initial points do not satisfy the '
                'critera necessary for convergence. Please specify different '
                'initial points, or increase FW_iter_limit')
        print('done.')

    def _check_solve(self, results, model_name):
        ''' Verify that the solver solved to optimality '''
        if (results.solver.status != pyo.SolverStatus.ok) or \
            (results.solver.termination_condition != pyo.TerminationCondition.optimal):
            print('Solve failed on model', model_name)
            print('Solver status:', results.solver.status)
            print('Termination conditions:', results.solver.termination_condition)
            raise RuntimeError()

    def _compute_dual_bound(self):
        ''' Compute the FWPH dual bound using self._local_bound from each rank
        '''
        send = np.array(self._local_bound)
        recv = np.array(0.)
        self.comms['ROOT'].Allreduce(
            [send, MPI.DOUBLE], [recv, MPI.DOUBLE], op=MPI.SUM)
        self._local_bound = recv

    def _conv_diff(self):
        ''' Perform the convergence check of Algorithm 3 in Boland et al. '''
        diff = 0.
        for name in self.local_subproblems.keys():
            mip = self.local_subproblems[name]
            arb_mip = self.local_scenarios[mip.scen_list[0]] \
                        if self.bundling else mip
            qp  = self.local_QP_subproblems[name]
            diff_s = 0.
            for (node_name, ix) in arb_mip._mpisppy_data.nonant_indices:
                qpx = qp.xr if self.bundling else qp.x
                diff_s += np.power(pyo.value(qpx[node_name,ix]) - 
                        pyo.value(arb_mip._mpisppy_model.xbars[node_name,ix]), 2)
            diff_s *= mip._mpisppy_probability
            diff += diff_s
        diff = np.array(diff)
        recv = np.array(0.)
        self.comms['ROOT'].Allreduce(
            [diff, MPI.DOUBLE], [recv, MPI.DOUBLE], op=MPI.SUM)
        return recv

    def _create_initial_points(self):
        pc = self.FW_options['point_creator']
        if ('point_creator_data' in self.FW_options.keys()):
            pd = self.FW_options['point_creator_data']
        else:
            pd = None
        self.local_initial_points = dict()
        for scenario_name in self.local_scenario_names:
            pts = pc(scenario_name, point_creator_data=pd)
            self.local_initial_points[scenario_name] = pts

    def _extract_objective(self, mip):
        ''' Extract the original part of the provided MIP's objective function
            (no dual or prox terms), and create a copy containing the QP
            variables in place of the MIP variables.

            Args:
                mip (Pyomo ConcreteModel): MIP model for a scenario or bundle.

            Returns:
                obj (Pyomo Objective): objective function extracted
                    from the MIP
                new (Pyomo Expression): expression from the MIP model
                    objective with the MIP variables replaced by QP variables.
                    Does not inculde dual or prox terms.

            Notes:
                Acts on either a single-scenario model or a bundle
        '''
        mip_to_qp = mip._mpisppy_data.mip_to_qp
        obj = find_active_objective(mip)
        repn = generate_standard_repn(obj.expr, quadratic=True)
        if len(repn.nonlinear_vars) > 0:
            raise ValueError("FWPH does not support models with nonlinear objective functions")
        linear_vars = [mip_to_qp[id(var)] for var in repn.linear_vars]
        new = LinearExpression(
            constant=repn.constant, linear_coefs=repn.linear_coefs, linear_vars=linear_vars
        )
        if repn.quadratic_vars:
            quadratic_vars = (
                (mip_to_qp[id(x)], mip_to_qp[id(y)]) for x,y in repn.quadratic_vars
            )
            new += pyo.quicksum(
                (coef*x*y for coef,(x,y) in zip(repn.quadratic_coefs, quadratic_vars))
            )
        return obj, new

    def _gather_weight_dict(self, strip_bundle_names=False):
        ''' Compute a double nested dictionary of the form

                weights[scenario name][variable name] = weight value

            for FWPH to return to the user.
        
            Notes:
                Must be called after the variables are swapped back.
        '''
        local_weights = dict()
        for (name, scenario) in self.local_scenarios.items():
            if (self.bundling and strip_bundle_names):
                scenario_weights = dict()
                for ndn_ix, var in scenario._mpisppy_data.nonant_indices.items():
                    rexp = '^' + scenario.name + '\.'
                    var_name = re.sub(rexp, '', var.name)
                    scenario_weights[var_name] = \
                                        scenario._mpisppy_model.W[ndn_ix].value
            else:
                scenario_weights = {nonant.name: scenario._mpisppy_model.W[ndn_ix].value
                        for ndn_ix, nonant in scenario._mpisppy_data.nonant_indices.items()}
            local_weights[name] = scenario_weights

        weights = self.comms['ROOT'].gather(local_weights, root=0)
        return weights

    def _get_leaf_vars(self, scenario):
        ''' This method simply needs to take an input scenario
            (pyo.ConcreteModel) and return a list of variable objects
            corresponding to the leaf node variables for that scenario.

            Functions by returning the complement of the set of
            non-anticipative variables.
        '''
        nonant_var_ids = [id(var) for node in scenario._mpisppy_node_list
                                  for var  in node.nonant_vardata_list]
        return [var for var in scenario.component_data_objects(pyo.Var)
                         if id(var) not in nonant_var_ids]

    def _get_xbars(self, strip_bundle_names=False):
        ''' Return the xbar vector if rank = 0 and None, otherwise
            (Consistent with _gather_weight_dict).

            Args:
                TODO
                
            Notes:
                Paralellism is not necessary since each rank already has an
                identical copy of xbar, provided by Compute_Xbar().

                Returned dictionary is indexed by variable name 
                (as provided by the user).

                Must be called after variables are swapped back (I think).
        '''
        if (self.cylinder_rank != 0):
            return None
        else:
            random_scenario_name = list(self.local_scenarios.keys())[0]
            scenario = self.local_scenarios[random_scenario_name]
            xbar_dict = {}
            for node in scenario._mpisppy_node_list:
                for (ix, var) in enumerate(node.nonant_vardata_list):
                    var_name = var.name
                    if (self.bundling and strip_bundle_names):
                        rexp = '^' + random_scenario_name + '\.'
                        var_name = re.sub(rexp, '', var_name)
                    xbar_dict[var_name] = scenario._mpisppy_model.xbars[node.name, ix].value
            return xbar_dict

    def _initialize_MIP_var_values(self):
        ''' Initialize the MIP variable values to the user-specified point.

            For now, arbitrarily choose the first point in each list.
        '''
        points = self.local_initial_points
        for (name, mip) in self.local_subproblems.items():
            pt = points[name][0] # Select the first point arbitrarily
            mip_vars = list(mip.component_data_objects(pyo.Var))
            for var in mip_vars:
                try:
                    var.set_value(pt[var.name])
                except KeyError as e:
                    raise KeyError('Found variable named' + var.name +
                        ' in model ' + name + ' not contained in the '
                        'specified initial point dictionary') from e

    def _initialize_QP_subproblems(self):
        ''' Instantiates the (convex) QP subproblems (eqn. (13) in the Boland
            paper) for each scenario. Does not create/attach an objective.

            Attachs a local_QP_subproblems dict to self. Keys are scenario
            names (or bundle names), values are Pyomo ConcreteModel objects
            corresponding to the QP subproblems. 

            QP subproblems are in their original form, without the x and y
            variables eliminated. Rationale: pre-solve will get this, easier
            bookkeeping (objective does not need to be changed at each inner
            iteration this way).
        '''
        self.local_QP_subproblems = dict()
        has_init_pts = hasattr(self, 'local_initial_points')
        for (name, model) in self.local_subproblems.items():
            if (self.bundling):
                xr_indices = model.ref_vars.keys()
                nonant_indices = model.nonant_vars.keys()
                leaf_indices = model.leaf_vars.keys()
                if (has_init_pts):
                    raise RuntimeError('Cannot currently specify '
                        'initial points while using bundles')
            else:
                nonant_indices = model._mpisppy_data.nonant_indices.keys()
                leaf_indices = model.leaf_vars.keys()

            ''' Convex comb. coefficients '''
            QP = pyo.ConcreteModel()
            QP.a = pyo.VarList(domain=pyo.NonNegativeReals)
            if (has_init_pts):
                for _ in range(len(self.local_initial_points[name])):
                    QP.a.add()
            else:
                QP.a.add() # Just one variable (1-based index!) to start

            ''' Other variables '''
            QP.x = pyo.Var(nonant_indices, within=pyo.Reals)
            QP.y = pyo.Var(leaf_indices, within=pyo.Reals)
            if (self.bundling):
                QP.xr = pyo.Var(xr_indices, within=pyo.Reals)

            ''' Non-anticipativity constraint '''
            if (self.bundling):
                def nonant_rule(m, scenario_name, node_name, ix):
                    return m.x[scenario_name, node_name, ix] == \
                            m.xr[node_name, ix]
                QP.na = pyo.Constraint(nonant_indices, rule=nonant_rule)
            
            ''' (x,y) constraints '''
            if (self.bundling):
                def x_rule(m, node_name, ix):
                    return -m.xr[node_name, ix] + m.a[1] * \
                            model.ref_vars[node_name, ix].value == 0
                def y_rule(m, scenario_name, node_name, ix):
                    return -m.y[scenario_name, node_name, ix] + m.a[1]\
                        * model.leaf_vars[scenario_name,node_name,ix].value == 0 
                QP.eqx = pyo.Constraint(xr_indices, rule=x_rule)
            else:
                if (has_init_pts):
                    pts = self.local_initial_points[name]
                    def x_rule(m, node_name, ix):
                        nm = model.nonant_vars[node_name, ix].name
                        return -m.x[node_name, ix] + \
                            pyo.quicksum(m.a[i+1] * pts[i][nm] 
                                for i in range(len(pts))) == 0
                    def y_rule(m, node_name, ix):
                        nm = model.leaf_vars[node_name, ix].name
                        return -m.y[node_name,ix] + \
                            pyo.quicksum(m.a[i+1] * pts[i][nm] 
                                for i in range(len(pts))) == 0
                else:
                    def x_rule(m, node_name, ix):
                        return -m.x[node_name, ix] + m.a[1] * \
                                model.nonant_vars[node_name, ix].value == 0
                    def y_rule(m, node_name, ix):
                        return -m.y[node_name,ix] + m.a[1] * \
                                model.leaf_vars['LEAF', ix].value == 0
                QP.eqx = pyo.Constraint(nonant_indices, rule=x_rule)

            QP.eqy = pyo.Constraint(leaf_indices, rule=y_rule)
            QP.sum_one = pyo.Constraint(expr=pyo.quicksum(QP.a.values())==1)

            QP._mpisppy_data = pyo.Block(name="For non-Pyomo mpi-sppy data")
            QP._mpisppy_model = pyo.Block(name="For mpi-sppy Pyomo additions to the scenario model")

            self.local_QP_subproblems[name] = QP
                
    def _initialize_QP_var_values(self):
        ''' Set the value of the QP variables to be equal to the values of the
            corresponding MIP variables.

            Notes:
                Must be called before _swap_nonant_vars()

                Must be called after _initialize_MIP_var_values(), if the user
                specifies initial sets of points. Otherwise, it must be called
                after Iter0().
        '''
        for name in self.local_subproblems.keys():
            mip = self.local_subproblems[name]
            qp  = self.local_QP_subproblems[name]

            for key in mip._mpisppy_model.x_indices:
                qp.x[key].set_value(mip.nonant_vars[key].value)
            for key in mip._mpisppy_model.y_indices:
                qp.y[key].set_value(mip.leaf_vars[key].value)

            # Set the non-anticipative reference variables if we're bundling
            if (self.bundling):
                arb_scenario = mip.scen_list[0]
                naix = self.local_scenarios[arb_scenario]._mpisppy_data.nonant_indices
                for (node_name, ix) in naix:
                    # Check that non-anticipativity is satisfied
                    # within the bundle (for debugging)
                    vals = [mip.nonant_vars[scenario_name, node_name, ix].value
                            for scenario_name in mip.scen_list]
                    assert(max(vals) - min(vals) < 1e-7)
                    qp.xr[node_name, ix].set_value(
                        mip.nonant_vars[arb_scenario, node_name, ix].value)

    def _is_timed_out(self):
        if (self.cylinder_rank == 0):
            time_elapsed = time.time() - self.t0
            status = 1 if (time_elapsed > self.FW_options['time_limit']) \
                       else 0
        else:
            status = None
        status = self.comms['ROOT'].bcast(status, root=0)
        return status != 0

    def _options_checks_fw(self):
        ''' Name                Boland notation (Algorithm 2)
            -------------------------------------------------
            FW_iter_limit       t_max
            FW_weight           alpha
            FW_conv_thresh      tau
        '''
        # 1. Check for required options
        reqd_options = ['FW_iter_limit', 'FW_weight', 'FW_conv_thresh',
                        'solvername']
        losers = [opt for opt in reqd_options if opt not in self.FW_options]
        if (len(losers) > 0):
            msg = "FW_options is missing the following key(s): " + \
                  ", ".join(losers)
            raise RuntimeError(msg)

        # 2. Check that bundles, pre-specified points and t_max play nice. This
        #    is only checked on rank 0, because that is where the initial
        #    points are supposed to be specified.
        use_bundles = ('bundles_per_rank' in self.options 
                        and self.options['bundles_per_rank'] > 0)
        t_max = self.FW_options['FW_iter_limit']
        specd_init_pts = 'point_creator' in self.FW_options.keys() and \
                         self.FW_options['point_creator'] is not None

        if (use_bundles and specd_init_pts):
            if (t_max == 1):
                raise RuntimeError('Cannot use bundles and specify initial '
                    'points with t_max=1 at the same time.')
            else:
                if (self.cylinder_rank == 0):
                    print('WARNING: Cannot specify initial points and use '
                        'bundles at the same time. Ignoring specified initial '
                        'points')
                # Remove specified initial points
                self.FW_options.pop('point_creator', None)

        if (t_max == 1 and not specd_init_pts):
            raise RuntimeError('FW_iter_limit set to 1. To ensure '
                'convergence, provide initial points, or increase '
                'FW_iter_limit')

        # 3. Check that the user did not specify the linearization of binary
        #    proximal terms (no binary variables allowed in FWPH QPs)
        if ('linearize_binary_proximal_terms' in self.options
            and self.options['linearize_binary_proximal_terms']):
            print('Warning: linearize_binary_proximal_terms cannot be used '
                  'with the FWPH algorithm. Ignoring...')
            self.options['linearize_binary_proximal_terms'] = False

        # 4. Provide a time limit of inf if the user did not specify
        if ('time_limit' not in self.FW_options.keys()):
            self.FW_options['time_limit'] = np.inf

    def _output(self, itr, bound, best_bound, diff, secs):
        if (self.cylinder_rank == 0 and self.vb):
            print('{itr:3d} {bound:12.4f} {best_bound:12.4f} {diff:12.4e} {secs:11.1f}s'.format(
                    itr=itr, bound=bound, best_bound=best_bound, 
                    diff=diff, secs=secs))
        if (self.cylinder_rank == 0 and 'save_file' in self.FW_options.keys()):
            fname = self.FW_options['save_file']
            with open(fname, 'a') as f:
                f.write('{itr:d},{bound:.16f},{best_bound:.16f},{diff:.16f},{secs:.16f}\n'.format(
                    itr=itr, bound=bound, best_bound=best_bound,
                    diff=diff, secs=secs))

    def _output_header(self):
        if (self.cylinder_rank == 0 and self.vb):
            print('itr {bound:>12s} {bb:>12s} {cd:>12s} {tm:>12s}'.format(
                    bound="bound", bb="best bound", cd="conv diff", tm="time"))
        if (self.cylinder_rank == 0 and 'save_file' in self.FW_options.keys()):
            fname = self.FW_options['save_file']
            with open(fname, 'a') as f:
                f.write('{itr:s},{bound:s},{bb:s},{diff:s},{secs:s}\n'.format(
                    itr="Iteration", bound="Bound", bb="Best bound",
                    diff="Error", secs="Time(s)"))

    def save_weights(self, fname):
        ''' Save the computed weights to the specified file.

            Notes:
                Handles parallelism--only writes one copy of the file.

                Rather "fast-and-loose", in that it doesn't enforce _when_ this
                function can be called.
        '''
        weights = self._gather_weight_dict(strip_bundle_names=self.bundling) # None if rank != 0
        if (self.cylinder_rank != 0):
            return
        with open(fname, 'w') as f:
            for block in weights:
                for (scenario_name, wts) in block.items():
                    for (var_name, weight_val) in wts.items():
                        row = '{sn},{vn},{wv:.16f}\n'.format(
                            sn=scenario_name, vn=var_name, wv=weight_val)
                        f.write(row)

    def save_xbars(self, fname):
        ''' Save the computed xbar to the specified file.

            Notes:
                Handles parallelism--only writes one copy of the file.

                Rather "fast-and-loose", in that it doesn't enforce _when_ this
                function can be called.
        '''
        if (self.cylinder_rank != 0):
            return
        xbars = self._get_xbars(strip_bundle_names=self.bundling) # None if rank != 0
        with open(fname, 'w') as f:
            for (var_name, xbs) in xbars.items():
                row = '{vn},{vv:.16f}\n'.format(vn=var_name, vv=xbs)
                f.write(row)

    def _set_MIP_solver_options(self):
        mip_opts = self.FW_options['mip_solver_options']
        if (len(mip_opts) > 0):
            for model in self.local_subproblems.values():
                for (key, option) in mip_opts.items():
                    model._solver_plugin.options[key] = option

    def _set_QP_objective(self):
        ''' Attach dual weights, objective function and solver to each QP.
        
            QP dual weights are initialized to the MIP dual weights.
        '''

        for name, mip in self.local_subproblems.items():
            QP = self.local_QP_subproblems[name]

            obj, new = self._extract_objective(mip)

            ## Finish setting up objective for QP
            if self.bundling:
                m_source = self.local_scenarios[mip.scen_list[0]]
                x_source = QP.xr
            else:
                m_source = mip
                x_source = QP.x

            QP._mpisppy_model.W = pyo.Param(
                m_source._mpisppy_data.nonant_indices.keys(), mutable=True, initialize=m_source._mpisppy_model.W
            )
            # rhos are attached to each scenario, not each bundle (should they be?)
            ph_term = pyo.quicksum((
                QP._mpisppy_model.W[nni] * x_source[nni] +
                (m_source._mpisppy_model.rho[nni] / 2.) * (x_source[nni] - m_source._mpisppy_model.xbars[nni]) * (x_source[nni] - m_source._mpisppy_model.xbars[nni])
                for nni in m_source._mpisppy_data.nonant_indices
            ))

            if obj.is_minimizing():
                QP.obj = pyo.Objective(expr=new+ph_term, sense=pyo.minimize)
            else:
                QP.obj = pyo.Objective(expr=-new+ph_term, sense=pyo.minimize)

            ''' Attach a solver with various options '''
            solver = pyo.SolverFactory(self.FW_options['solvername'])
            if sputils.is_persistent(solver):
                solver.set_instance(QP)
            if 'qp_solver_options' in self.FW_options:
                qp_opts = self.FW_options['qp_solver_options']
                if qp_opts:
                    for (key, option) in qp_opts.items():
                        solver.options[key] = option

            self.local_QP_subproblems[name]._QP_solver_plugin = solver

    def _swap_nonant_vars(self):
        ''' Change the pointers in
            scenario._mpisppy_node_list[i].nonant_vardata_list
            to point to the QP variables, rather than the MIP variables.

            Notes:
                When computing xBar and updating the weights in the outer
                iteration, the values of the x variables are pulled from
                scenario._mpisppy_node_list[i].nonant_vardata_list. In the FWPH
                algorithm, xBar should be computed using the QP values, not the
                MIP values (like in normal PH).

                Reruns SPBase._attach_nonant_indices so that the scenario 
                _nonant_indices dictionary has the correct variable pointers
                
                Updates nonant_vardata_list but NOT nonant_list.
        '''
        for (name, model) in self.local_subproblems.items():
            scens = model.scen_list if self.bundling else [name]
            for scenario_name in scens:
                scenario = self.local_scenarios[scenario_name]
                num_nonant_vars = scenario._mpisppy_data.nlens
                node_list = scenario._mpisppy_node_list
                for node in node_list:
                    node.nonant_vardata_list = [
                        self.local_QP_subproblems[name].xr[node.name,i]
                        if self.bundling else
                        self.local_QP_subproblems[name].x[node.name,i]
                        for i in range(num_nonant_vars[node.name])]
        self._attach_nonant_indices()

    def _swap_nonant_vars_back(self):
        ''' Swap variables back, in case they're needed somewhere else.
        '''
        for (name, model) in self.local_subproblems.items():
            if (self.bundling):
                EF = self.local_subproblems[name]
                for scenario_name in EF.scen_list:
                    scenario = self.local_scenarios[scenario_name]
                    num_nonant_vars = scenario._mpisppy_data.nlens
                    for node in scenario._mpisppy_node_list:
                        node.nonant_vardata_list = [
                            EF.nonant_vars[scenario_name,node.name,ix]
                            for ix in range(num_nonant_vars[node.name])]
            else:
                scenario = self.local_scenarios[name]
                num_nonant_vars = scenario._mpisppy_data.nlens
                for node in scenario._mpisppy_node_list:
                    node.nonant_vardata_list = [
                        scenario.nonant_vars[node.name,ix]
                        for ix in range(num_nonant_vars[node.name])]
        self._attach_nonant_indices()

if __name__=='__main__':
    print('fwph.py has no main()')



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# Provide some test for confidence intervals
"""

"""

import os
import tempfile
import numpy as np
import unittest
import subprocess


import mpi4py.MPI as mpi

from mpisppy.tests.test_utils import get_solver, round_pos_sig
import mpisppy.tests.examples.farmer as farmer


import mpisppy.confidence_intervals.mmw_ci as MMWci
import mpisppy.utils.amalgomator as ama
from mpisppy.utils.xhat_eval import Xhat_Eval
import mpisppy.confidence_intervals.seqsampling as seqsampling

fullcomm = mpi.COMM_WORLD
global_rank = fullcomm.Get_rank()

__version__ = 0.2

solver_available,solvername, persistent_available, persistentsolvername= get_solver()
module_dir = os.path.dirname(os.path.abspath(__file__))

def _get_base_options():
    options = { "EF_solver_name": solvername,
                     "use_integer": False,
                     "crops_multiplier": 1,
                     'num_scens': 12,
                     'start': 0}
    Baseoptions =  {"num_batches": 2,
                     "batch_size": 10,
                     "opt":options}
    scenario_creator_kwargs = farmer.kw_creator(options)
    Baseoptions['kwargs'] = scenario_creator_kwargs
    return Baseoptions

def _get_xhatEval_options():
    options = {"iter0_solver_options": None,
             "iterk_solver_options": None,
             "display_timing": False,
             "solvername": solvername,
             "verbose": False,
             "solver_options":None}
    return options


refmodelname ="mpisppy.tests.examples.farmer"
#*****************************************************************************
class Test_MMW_farmer(unittest.TestCase):
    """ Test the MMWci code using farmer."""

    def setUp(self):
        self.xhat = {'ROOT': np.array([74.0,245.0,181.0])}
        tmpxhat = tempfile.mkstemp(prefix="xhat",suffix=".npy")
        self.xhat_path =  tmpxhat[1]#create an empty .npy file
        MMWci.write_xhat(self.xhat,self.xhat_path)

    def tearDown(self):
        os.remove(self.xhat_path)

    def test_MMW_constructor(self):
        options = _get_base_options()
        xhat = MMWci.read_xhat(self.xhat_path)

        MMW = MMWci.MMWConfidenceIntervals(refmodelname,
                          options['opt'],
                          xhat,
                          options['num_batches'])
    
    def test_xhat_read_write(self):
        path = tempfile.mkstemp(prefix="xhat",suffix=".npy")[1]
        MMWci.write_xhat(self.xhat,path=path)
        x = MMWci.read_xhat(path, delete_file=True)
        self.assertEqual(list(x['ROOT']), list(self.xhat['ROOT']))
    
    def test_xhat_read_write_txt(self):
        path = tempfile.mkstemp(prefix="xhat",suffix=".npy")[1]
        MMWci.writetxt_xhat(self.xhat,path=path)
        x = MMWci.readtxt_xhat(path, delete_file=True)
        self.assertEqual(list(x['ROOT']), list(self.xhat['ROOT']))
    
    def test_ama_creator(self):
        options = _get_base_options()
        ama_options = {"EF-2stage": True,}
        ama_options.update(options['opt'])
        ama_object = ama.from_module(refmodelname,
                                     options=ama_options,
                                     use_command_line=False)   
        
    def test_seqsampling_creator(self):
            optionsBM = {'h':0.2,
                         'hprime':0.015, 
                         'eps':0.5, 
                         'epsprime':0.4, 
                         "p":2,
                         "q":1.2,
                         "solvername":solvername,
                         "xhat_gen_options": {"crops_multiplier": 3},
                         "crops_multiplier": 3,
                         }
            seqsampling.SeqSampling("mpisppy.tests.examples.farmer",
                                    seqsampling.xhat_generator_farmer,
                                    optionsBM,
                                    stochastic_sampling=False,
                                    stopping_criterion="BM",
                                    )

    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_ama_running(self):
        options = _get_base_options()
        ama_options = {"EF-2stage": True}
        ama_options.update(options['opt'])
        ama_object = ama.from_module(refmodelname,
                                     ama_options, use_command_line=False)
        ama_object.run()
        obj = round_pos_sig(ama_object.EF_Obj,2)
        self.assertEqual(obj, -130000)
    

    @unittest.skipIf(not solver_available,
                     "no solver is available")      
    def test_xhat_eval_creator(self):
        options = _get_xhatEval_options()
        
        MMW_options = _get_base_options()
        scenario_creator_kwargs = MMW_options['kwargs']
        scenario_creator_kwargs['num_scens'] = MMW_options['batch_size']
        ev = Xhat_Eval(options,
                       farmer.scenario_names_creator(100),
                       farmer.scenario_creator,
                       scenario_denouement=None,
                       scenario_creator_kwargs=scenario_creator_kwargs
                       )
        
    @unittest.skipIf(not solver_available,
                     "no solver is available")      
    def test_xhat_eval_evaluate(self):
        options = _get_xhatEval_options()
        MMW_options = _get_base_options()
        scenario_creator_kwargs = MMW_options['kwargs']
        scenario_creator_kwargs['num_scens'] = MMW_options['batch_size']
        ev = Xhat_Eval(options,
                   farmer.scenario_names_creator(100),
                   farmer.scenario_creator,
                   scenario_denouement=None,
                   scenario_creator_kwargs=scenario_creator_kwargs
                   )
        
        xhat = MMWci.read_xhat(self.xhat_path)
        obj = round_pos_sig(ev.evaluate(xhat),2)
        self.assertEqual(obj, -1300000.0)
 
    @unittest.skipIf(not solver_available,
                     "no solver is available")  
    def test_xhat_eval_evaluate_one(self):
        options = _get_xhatEval_options()
        MMW_options = _get_base_options()
        xhat = MMWci.read_xhat(self.xhat_path)
        scenario_creator_kwargs = MMW_options['kwargs']
        scenario_creator_kwargs['num_scens'] = MMW_options['batch_size']
        scenario_names = farmer.scenario_names_creator(100)
        ev = Xhat_Eval(options,
                   scenario_names,
                   farmer.scenario_creator,
                   scenario_denouement=None,
                   scenario_creator_kwargs=scenario_creator_kwargs
                   )
        k = scenario_names[0]
        obj = ev.evaluate_one(xhat,k,ev.local_scenarios[k])
        obj = round_pos_sig(obj,2)
        self.assertEqual(obj, -48000.0)
      
    @unittest.skipIf(not solver_available,
                     "no solver is available")  
    def test_MMW_running(self):
        options = _get_base_options()
        xhat = MMWci.read_xhat(self.xhat_path)
        MMW = MMWci.MMWConfidenceIntervals(refmodelname,
                                        options['opt'],
                                        xhat,
                                        options['num_batches'])
        r = MMW.run() 
        s = round_pos_sig(r['std'],2)
        bound = round_pos_sig(r['gap_inner_bound'],2)
        self.assertEqual((s,bound), (43.0,280.0))
   
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_mmwci_main(self):
        current_dir = os.getcwd()
        os.chdir(os.path.join(module_dir,"..","confidence_intervals"))
        runstring = "python mmw_ci.py --num-scens=3  --MMW-num-batches=3 --MMW-batch-size=3 " +\
                    "--EF-solver-name="+solvername
        res = str(subprocess.check_output(runstring, shell=True))
        os.chdir(current_dir)
        self.assertIn("1050.77",res)
    
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_gap_estimators(self):
        scenario_names = farmer.scenario_names_creator(50,start=1000)
        G,s = seqsampling.gap_estimators(self.xhat,
                                         solvername,
                                         scenario_names,
                                         farmer.scenario_creator)
        G,s = round_pos_sig(G,3),round_pos_sig(s,3)
        self.assertEqual((G,s), (110.0,426.0))
        
    @unittest.skipIf(not solver_available,
                     "no solver is available")
    def test_seqsampling_running(self):
            #We need a very small instance for testing on GitHub.
            optionsBM = {'h':1.75,
                         'hprime':0.5, 
                         'eps':0.2, 
                         'epsprime':0.1, 
                         "p":0.1,
                         "q":1.2,
                         "solvername":solvername,
                         "xhat_gen_options": {"crops_multiplier": 3},
                         "crops_multiplier": 3,
                         }
            seq_pb = seqsampling.SeqSampling("mpisppy.tests.examples.farmer",
                                            seqsampling.xhat_generator_farmer,
                                            optionsBM,
                                            stochastic_sampling=False,
                                            stopping_criterion="BM",
                                            )
            x = seq_pb.run(maxit=50)
            T = x['T']
            ub = round_pos_sig(x['CI'][1],2)
            self.assertEqual((T,ub), (5,14000.0))
            
if __name__ == '__main__':
    unittest.main()
    



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.

# Adapted from the PySP extension adaptive_rho_converger, authored by Gabriel Hackebeil

import math
import mpisppy.extensions.extension

import numpy as np
import mpi4py.MPI as MPI

_norm_rho_defaults = { 'convergence_tolerance' : 1e-4,
                           'rho_decrease_multiplier' : 2.0,
                           'rho_increase_multiplier' : 2.0,
                           'primal_dual_difference_factor' : 100.,
                           'iterations_converged_before_decrease' : 0,
                           'rho_converged_decrease_multiplier' : 1.1,
                           'rho_update_stop_iterations' : None,
                           'verbose' : False,
}

_attr_to_option_name_map = {
    '_tol': 'convergence_tolerance',
    '_rho_decrease' : 'rho_decrease_multiplier',
    '_rho_increase' : 'rho_increase_multiplier',
    '_primal_dual_difference_factor' : 'primal_dual_difference_factor',
    '_required_converged_before_decrease' : 'iterations_converged_before_decrease',
    '_rho_converged_residual_decrease' : 'rho_converged_decrease_multiplier',
    '_stop_iter_rho_update' : 'rho_update_stop_iterations',
    '_verbose' : 'verbose',
}

class NormRhoUpdater(mpisppy.extensions.extension.Extension):

    def __init__(self, ph):

        self.ph = ph
        self.norm_rho_options = \
            ph.options['norm_rho_options'] if 'norm_rho_options' in ph.options else dict()

        self._set_options()
        self._prev_avg = None

    def _set_options(self):
        options = self.norm_rho_options
        for attr_name, opt_name in _attr_to_option_name_map.items():
            setattr(self, attr_name, options[opt_name] if opt_name in options else _norm_rho_defaults[opt_name])

    def _snapshot_avg(self, ph):
        scenario = ph.local_scenarios[ph.local_scenario_names[0]]
        self._prev_avg = { ndn_i : xbar.value for ndn_i, xbar in scenario._mpisppy_model.xbars.items() }

    def _compute_primal_residual_norm(self, ph):

        local_nodenames = []
        local_primal_residuals = {}
        global_primal_residuals = {}

        for k,s in ph.local_scenarios.items():
            nlens = s._mpisppy_data.nlens        
            for node in s._mpisppy_node_list:
                if node.name not in local_nodenames:

                    ndn = node.name
                    local_nodenames.append(ndn)
                    nlen = nlens[ndn]

                    local_primal_residuals[ndn] = np.zeros(nlen, dtype='d')
                    global_primal_residuals[ndn] = np.zeros(nlen, dtype='d')

        for k,s in ph.local_scenarios.items():
            nlens = s._mpisppy_data.nlens
            xbars = s._mpisppy_model.xbars
            for node in s._mpisppy_node_list:
                ndn = node.name
                primal_residuals = local_primal_residuals[ndn]

                unweighted_primal_residuals = \
                        np.fromiter((abs(v._value - xbars[ndn,i]._value) for i,v in enumerate(node.nonant_vardata_list)),
                                    dtype='d', count=nlens[ndn] )
                primal_residuals += s._mpisppy_probability * unweighted_primal_residuals

        for nodename in local_nodenames:
            ph.comms[nodename].Allreduce(
                [local_primal_residuals[nodename], MPI.DOUBLE],
                [global_primal_residuals[nodename], MPI.DOUBLE],
                op=MPI.SUM)

        primal_resid = {}
        for ndn, global_primal_resid in global_primal_residuals.items():
            for i, v in enumerate(global_primal_resid):
                primal_resid[ndn,i] = v

        return primal_resid

    def _compute_dual_residual_norm(self, ph):
        dual_resid = {}
        s = ph.local_scenarios[ph.local_scenario_names[0]]
        for ndn_i in s._mpisppy_data.nonant_indices:
            dual_resid[ndn_i] = s._mpisppy_model.rho[ndn_i].value * \
                                math.fabs( s._mpisppy_model.xbars[ndn_i].value - self._prev_avg[ndn_i] )
        return dual_resid

    def pre_iter0(self):
        pass

    def post_iter0(self):
        pass

    def miditer(self):

        ph = self.ph
        ph_iter = ph._PHIter
        if self._stop_iter_rho_update is not None and \
                (ph_iter > self._stop_iter_rho_update):
            return
        if self._prev_avg is None:
            self._snapshot_avg(ph)
        else:
            primal_residuals = self._compute_primal_residual_norm(ph)
            dual_residuals = self._compute_dual_residual_norm(ph)
            self._snapshot_avg(ph)
            primal_dual_difference_factor = self._primal_dual_difference_factor
            first = True
            first_scenario = True
            for s in ph.local_scenarios.values():
                for ndn_i, rho in s._mpisppy_model.rho.items():
                    primal_resid = primal_residuals[ndn_i]
                    dual_resid = dual_residuals[ndn_i]

                    action = None
                    if (primal_resid > primal_dual_difference_factor*dual_resid) and (primal_resid > self._tol):
                        rho._value *= self._rho_increase
                        action = "Increasing"
                    elif (dual_resid > primal_dual_difference_factor*primal_resid) and (dual_resid > self._tol):
                        if ph_iter >= self._required_converged_before_decrease:
                            rho._value /= self._rho_decrease
                            action = "Decreasing"
                    elif (primal_resid < self._tol) and (dual_resid < self._tol):
                        rho._value /= self._rho_converged_residual_decrease
                        action = "Converged, Decreasing"
                    if self._verbose and ph.cylinder_rank == 0 and action is not None:
                        if first:
                            first = False
                            first_line = ("Updating rho values:\n%21s %40s %16s %16s %16s"
                                          % ("Action",
                                             "Variable",
                                             "Primal Residual",
                                             "Dual Residual",
                                             "New Rho"))
                            print(first_line)
                        if first_scenario:
                            print("%21s %40s %16g %16g %16g"
                                  % (action, s._mpisppy_data.nonant_indices[ndn_i].name,
                                     primal_resid, dual_resid, rho.value))
                first_scenario = False

    def enditer(self):
        pass

    def post_everything(self):
        pass



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# updated April 2020
import mpisppy.cylinders.spoke as spoke
from mpisppy.extensions.xhatlooper import XhatLooper
from mpisppy.utils.xhat_eval import Xhat_Eval
import logging
import mpisppy.log

# Could also pass, e.g., sys.stdout instead of a filename
mpisppy.log.setup_logger("mpisppy.cylinders.xhatlooper_bounder",
                         "xhatlp.log",
                         level=logging.CRITICAL)                         
logger = logging.getLogger("mpisppy.cylinders.xhatlooper_bounder")


class XhatLooperInnerBound(spoke.InnerBoundNonantSpoke):

    converger_spoke_char = 'X'

    def xhatlooper_prep(self):
        verbose = self.opt.options['verbose']
        if "bundles_per_rank" in self.opt.options\
           and self.opt.options["bundles_per_rank"] != 0:
            raise RuntimeError("xhat spokes cannot have bundles (yet)")

        if not isinstance(self.opt, Xhat_Eval):
            raise RuntimeError("XhatShuffleInnerBound must be used with Xhat_Eval.")

        xhatter = XhatLooper(self.opt)

        ### begin iter0 stuff
        xhatter.pre_iter0()
        self.opt._save_original_nonants()

        teeme = False
        if "tee-rank0-solves" in self.opt.options:
            teeme = self.opt.options['tee-rank0-solves']

        self.opt.solve_loop(
            solver_options=self.opt.current_solver_options,
            dtiming=False,
            gripe=True,
            tee=teeme,
            verbose=verbose
        )
        self.opt._update_E1()  # Apologies for doing this after the solves...
        if abs(1 - self.opt.E1) > self.opt.E1_tolerance:
            if self.opt.cylinder_rank == 0:
                print("ERROR")
                print("Total probability of scenarios was ", self.opt.E1)
                print("E1_tolerance = ", self.opt.E1_tolerance)
            quit()
        infeasP = self.opt.infeas_prob()
        if infeasP != 0.:
            if self.opt.cylinder_rank == 0:
                print("ERROR")
                print("Infeasibility detected; E_infeas, E1=", infeasP, self.opt.E1)
            quit()
        ### end iter0 stuff

        xhatter.post_iter0()
        self.opt._save_nonants() # make the cache

        return xhatter

    def main(self):
        verbose = self.opt.options["verbose"] # typing aid  
        logger.debug(f"Entering main on xhatlooper spoke rank {self.global_rank}")

        xhatter = self.xhatlooper_prep()

        scen_limit = self.opt.options['xhat_looper_options']['scen_limit']

        xh_iter = 1
        while not self.got_kill_signal():
            if (xh_iter-1) % 10000 == 0:
                logger.debug(f'   Xhatlooper loop iter={xh_iter} on rank {self.global_rank}')
                logger.debug(f'   Xhatlooper got from opt on rank {self.global_rank}')

            if self.new_nonants:
                logger.debug(f'   *Xhatlooper loop iter={xh_iter}')
                logger.debug(f'   *got a new one! on rank {self.global_rank}')
                logger.debug(f'   *localnonants={str(self.localnonants)}')

                self.opt._put_nonant_cache(self.localnonants)
                self.opt._restore_nonants()
                upperbound, srcsname = xhatter.xhat_looper(scen_limit=scen_limit, restore_nonants=False)

                # send a bound to the opt companion
                self.update_if_improving(upperbound)
            xh_iter += 1



#ReferenceModel for full set of scenarios for APL1P; May 2021


import pyomo.environ as pyo
import numpy as np
import mpisppy.scenario_tree as scenario_tree
import mpisppy.utils.sputils as sputils
import mpisppy.utils.amalgomator as amalgomator

# Use this random stream:
apl1pstream = np.random.RandomState()

def APL1P_model_creator(seed):
    apl1pstream.seed(seed)
    random_array = apl1pstream.rand(6) #We only use 5 random numbers
    
    #
    # Model
    #
    
    model = pyo.ConcreteModel()
    
    #
    # Parameters
    #
    
    # generator
    model.G = [1,2]
    
    # Demand level
    model.DL = [1,2,3]
    
    # Availability
    avail_outcome = ([1.,0.9,0.5,0.1],[1.,0.9,0.7,0.1,0.0])
    avail_probability = ([0.2,0.3,0.4,0.1],[0.1,0.2,0.5,0.1,0.1])
    avail_cumprob = (np.cumsum(avail_probability[0]),np.cumsum(avail_probability[1]))
    assert(max(avail_cumprob[0])==1.0 and max(avail_cumprob[1])==1.0)
    def availability_init(m,g):
        rd = random_array[g]
        i = np.searchsorted(avail_cumprob[g-1],rd)
        return avail_outcome[g-1][i]
    
    model.Availability = pyo.Param(model.G, within=pyo.NonNegativeReals,
                               initialize=availability_init)
    
    # Min Capacity
    cmin_init = 1000
    model.Cmin = pyo.Param(model.G, within=pyo.NonNegativeReals, initialize=cmin_init)
    
    
    # Investment, aka Capacity costs
    invest = np.array([.4,.25])
    def investment_init(m,g):
        return(invest[g-1])

    model.Investment = pyo.Param(model.G, within=pyo.NonNegativeReals,
                             initialize=investment_init)
    
    # Operating Cost
    op_cost = np.array([[.43,.2,.05],[.87,.4,.1]])
    def operatingcost_init(m,g,dl):
        return(op_cost[g-1,dl-1])
    
    model.OperatingCost = pyo.Param(model.G, model.DL, within=pyo.NonNegativeReals,
                                initialize = operatingcost_init)
    
    # Demand
    demand_outcome = [900,1000,1100,1200]
    demand_prob = [.15,.45,.25,.15]
    demand_cumprob = np.cumsum(demand_prob)
    assert(max(demand_cumprob == 1.0))
    def demand_init(m,dl):
        rd = random_array[2+dl]
        i = np.searchsorted(demand_cumprob,rd)
        return demand_outcome[i]
    
    model.Demand = pyo.Param(model.DL, within=pyo.NonNegativeReals,
                         initialize=demand_init)
    
    # Cost of unserved demand
    unserved_cost =1.0
    model.CostUnservedDemand = pyo.Param(model.DL, within=pyo.NonNegativeReals,
                                     initialize=unserved_cost)
    
    #
    # Variables
    #
    
    # Capacity of generators
    model.CapacityGenerators = pyo.Var(model.G, domain=pyo.NonNegativeReals)
    
    # Operation level
    model.OperationLevel = pyo.Var(model.G, model.DL, domain=pyo.NonNegativeReals)
    
    # Unserved demand
    model.UnservedDemand = pyo.Var(model.DL, domain=pyo.NonNegativeReals)
    
    
    #
    # Constraints
    #
    
    # Minimum capacity
    def MinimumCapacity_rule(model, g):
        return model.CapacityGenerators[g] >= model.Cmin[g]
    
    model.MinimumCapacity = pyo.Constraint(model.G, rule=MinimumCapacity_rule)
    
    # Maximum operating level
    def MaximumOperating_rule(model, g):
        return sum(model.OperationLevel[g, dl] for dl in model.DL) <= model.Availability[g] * model.CapacityGenerators[g]
    
    model.MaximumOperating = pyo.Constraint(model.G, rule=MaximumOperating_rule)
    
    # Satisfy demand
    def SatisfyDemand_rule(model, dl):
        return sum(model.OperationLevel[g, dl] for g in model.G) + model.UnservedDemand[dl] >= model.Demand[dl]
    
    model.SatisfyDemand = pyo.Constraint(model.DL, rule=SatisfyDemand_rule)
    
    #
    # Stage-specific cost computations
    #
    
    def ComputeFirstStageCost_rule(model):
        return sum(model.Investment[g] * model.CapacityGenerators[g] for g in model.G)
    
    model.FirstStageCost = pyo.Expression(rule=ComputeFirstStageCost_rule)
    
    
    def ComputeSecondStageCost_rule(model):
        expr = sum(
            model.OperatingCost[g, dl] * model.OperationLevel[g, dl] for g in model.G for dl in model.DL) + sum(
            model.CostUnservedDemand[dl] * model.UnservedDemand[dl] for dl in model.DL)
        return expr

    model.SecondStageCost = pyo.Expression(rule=ComputeSecondStageCost_rule)
    
    
    def total_cost_rule(model):
        return model.FirstStageCost + model.SecondStageCost
    
    model.Total_Cost_Objective = pyo.Objective(rule=total_cost_rule, sense=pyo.minimize)

    return(model)


def scenario_creator(sname, num_scens=None):
    scennum   = sputils.extract_num(sname)
    model = APL1P_model_creator(scennum)
    
    # Create the list of nodes associated with the scenario (for two stage,
    # there is only one node associated with the scenario--leaf nodes are
    # ignored).
    model._mpisppy_node_list = [
        scenario_tree.ScenarioNode(
            name="ROOT",
            cond_prob=1.0,
            stage=1,
            cost_expression=model.Total_Cost_Objective,
            scen_name_list=None, # Deprecated?
            nonant_list=[model.CapacityGenerators], 
            scen_model=model,
        )
    ]
    
    #Add the probability of the scenario
    if num_scens is not None :
        model._mpisppy_probability = 1/num_scens
    
    return(model)


#=========
def scenario_names_creator(num_scens,start=None):
    # (only for Amalgomator): return the full list of num_scens scenario names
    # if start!=None, the list starts with the 'start' labeled scenario
    if (start is None) :
        start=0
    return [f"scen{i}" for i in range(start,start+num_scens)]
        

#=========
def inparser_adder(inparser):
    # (only for Amalgomator): add command options unique to apl1p
    pass


#=========
def kw_creator(options):
    # (only for Amalgomator): linked to the scenario_creator and inparser_adder
    kwargs = {"num_scens" : options['num_scens'] if 'num_scens' in options else None,
              }
    return kwargs


#============================
def scenario_denouement(rank, scenario_name, scenario):
    pass


#============================
def xhat_generator_apl1p(scenario_names, solvername="gurobi", solver_options=None):
    '''
    For sequential sampling.
    Takes scenario names as input and provide the best solution for the 
        approximate problem associated with the scenarios.
    Parameters
    ----------
    scenario_names: int
        Names of the scenario we use
    solvername: str, optional
        Name of the solver used. The default is "gurobi"
    solver_options: dict, optional
        Solving options. The default is None.

    Returns
    -------
    xhat: str
        A generated xhat, solution to the approximate problem induced by scenario_names.

    '''
    num_scens = len(scenario_names)
    
    ama_options = { "EF-2stage": True,
                    "EF_solver_name": solvername,
                    "EF_solver_options": solver_options,
                    "num_scens": num_scens,
                    "_mpisppy_probability": 1/num_scens,
                    }
    #We use from_module to build easily an Amalgomator object
    ama = amalgomator.from_module("mpisppy.tests.examples.apl1p",
                                  ama_options,use_command_line=False)
    #Correcting the building by putting the right scenarios.
    ama.scenario_names = scenario_names
    ama.run()
    
    # get the xhat
    xhat = sputils.nonant_cache_from_ef(ama.ef)

    return xhat

if __name__ == "__main__":
    #An example of sequential sampling for the APL1P model
    from mpisppy.confidence_intervals.seqsampling import SeqSampling
    optionsFSP = {'eps': 5.0,
                  'solvername': "gurobi_direct",
                  "c0":50,}
    apl1p_pb = SeqSampling("mpisppy.tests.examples.apl1p",
                            xhat_generator_apl1p, 
                            optionsFSP,
                            stopping_criterion="BPL",
                            stochastic_sampling=False)

    res = apl1p_pb.run()
    print(res)


# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
import mpisppy.cylinders.spoke

class LagrangianOuterBound(mpisppy.cylinders.spoke.OuterBoundWSpoke):

    converger_spoke_char = 'L'

    def lagrangian_prep(self):
        verbose = self.opt.options['verbose']
        # Split up PH_Prep? Prox option is important for APH.
        # Seems like we shouldn't need the Lagrangian stuff, so attach_prox=False
        # Scenarios are created here
        self.opt.PH_Prep(attach_prox=False)
        self.opt._reenable_W()
        self.opt.subproblem_creation(verbose)
        self.opt._create_solvers()

    def lagrangian(self):
        verbose = self.opt.options['verbose']
        # This is sort of a hack, but might help folks:
        if "ipopt" in self.opt.options["solvername"]:
            print("\n WARNING: An ipopt solver will not give outer bounds\n")
        teeme = False
        if "tee-rank0-solves" in self.opt.options:
            teeme = self.opt.options['tee-rank0-solves']

        self.opt.solve_loop(
            solver_options=self.opt.current_solver_options,
            dtiming=False,
            gripe=True,
            tee=teeme,
            verbose=verbose
        )
        ''' DTM (dlw edits): This is where PHBase Iter0 checks for scenario
            probabilities that don't sum to one and infeasibility and
            will send a kill signal if needed. For now we are relying
            on the fact that the OPT thread is solving the same
            models, and hence would detect both of those things on its
            own--the Lagrangian spoke doesn't need to check again.  '''

        # Compute the resulting bound, checking to be sure
        # the weights came from the same PH iteration
        serial_number = self.get_serial_number()
        bound, extra_sums  = self.opt.Ebound(verbose, extra_sum_terms=[serial_number])
        serial_number_sum = int(round(extra_sums[0]))

        total = int(self.cylinder_comm.Get_size())*serial_number
        if total == serial_number_sum:
            return bound
        elif self.cylinder_rank == 0:
            print("WARNING: Lagrangian spokes out of snyc, consider changing mpisppy.cylinders.SPOKE_SLEEP_TIME")
        return None

    def _set_weights_and_solve(self):
        self.opt.W_from_flat_list(self.localWs) # Sets the weights
        return self.lagrangian()

    def main(self):
        # The rho_setter should be attached to the opt object
        rho_setter = None
        if hasattr(self.opt, 'rho_setter'):
            rho_setter = self.opt.rho_setter

        self.lagrangian_prep()

        self.trivial_bound = self.lagrangian()

        self.bound = self.trivial_bound

        dk_iter = 1
        while not self.got_kill_signal():
            if self.new_Ws:
                bound = self._set_weights_and_solve()
                if bound is not None:
                    self.bound = bound
            dk_iter += 1

    def finalize(self):
        '''
        Do one final lagrangian pass with the final
        PH weights. Useful for when PH convergence
        and/or iteration limit is the cause of termination
        '''
        self.final_bound = self._set_weights_and_solve()
        self.bound = self.final_bound
        return self.final_bound



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
''' Driver script to test the w/xbar read/write extensions using UC 
'''

import mpisppy.tests.examples.uc.uc_funcs as uc_funcs
# this version is locked to three scenarios

from mpisppy.utils.wxbarwriter import WXBarWriter
from mpisppy.utils.wxbarreader import WXBarReader
import os

from mpisppy.opt.ph import PH

def nonsense(arg1, arg2, arg3): # Empty scenario_denouement
    pass

def read_test():
    scen_count          = 3
    scenario_creator    = uc_funcs.pysp2_callback
    scenario_denouement = nonsense
    scenario_rhosetter  = uc_funcs.scenario_rhos

    PH_options = {
        'solvername': 'gurobi',
        'PHIterLimit': 2,
        'defaultPHrho': 1,
        'convthresh': 1e-6,
        'verbose': False,
        'display_timing': False,
        'display_progress': False,
        'iter0_solver_options': dict(),
        'iterk_solver_options': dict(),
        'init_W_fname': 'david/weights.csv', # Path to the weight files
        'init_separate_W_files': False,
        'init_Xbar_fname': 'david/somexbars.csv',
        'extensions':WXBarReader,
        'rho_setter':scenario_rhosetter,
    }

    names = ['Scenario' + str(i+1) for i in range(scen_count)]

    ph = PH(PH_options, names, scenario_creator, scenario_denouement)

    conv, obj, bound = ph.ph_main()

def write_test():
    scen_count          = 3
    scenario_creator    = uc_funcs.pysp2_callback
    scenario_denouement = nonsense
    scenario_rhosetter  = uc_funcs.scenario_rhos

    PH_options = {
        'solvername': 'gurobi',
        'PHIterLimit': 2,
        'defaultPHrho': 1,
        'convthresh': 1e-6,
        'verbose': False,
        'display_timing': False,
        'display_progress': False,
        'iter0_solver_options': dict(),
        'iterk_solver_options': dict(),
        'W_fname': 'david/weights.csv',
        'separate_W_files': False,
        'Xbar_fname': 'somexbars.csv',
        'extensions':WXBarReader,
        'rho_setter':scenario_rhosetter,
    }

    names = ['Scenario' + str(i+1) for i in range(scen_count)]

    ph = PH(PH_options, names, scenario_creator, scenario_denouement)

    conv, obj, bound = ph.ph_main()

if __name__=='__main__':
    read_test()



# Copyright 2020 by B. Knueven, D. Mildebrath, C. Muir, J-P Watson, and D.L. Woodruff
# This software is distributed under the 3-clause BSD License.
# APH
"""
TBD: dlw june 2020 look at this code in phbase:
            if spcomm is not None: 
                spcomm.sync_with_spokes()
                if spcomm.is_converged():
                    break    

"""


import numpy as np
import math
import collections
import time
import logging
import mpi4py
import mpi4py.MPI as mpi
import pyomo.environ as pyo
from pyomo.opt import SolverFactory, SolverStatus
from mpisppy.utils.sputils import find_active_objective
import mpisppy.utils.listener_util.listener_util as listener_util
import mpisppy.phbase as ph_base
import mpisppy.utils.sputils as sputils

fullcomm = mpi.COMM_WORLD
global_rank = fullcomm.Get_rank()

logging.basicConfig(level=logging.CRITICAL, # level=logging.CRITICAL, DEBUG
            format='(%(threadName)-10s) %(message)s',
            )


"""
Delete this comment block; dlw May 2019
- deal with "waiting out" a negative tau
"""

""" APH started by DLW, March 2019.
Based on "Algorithm 2: Asynchronous projective hedging (APH) -- Algorithm 1
specialize to the setup S1-S4" from 
"Asynchronous Projective Hedging for Stochastic Programming" 
http://www.optimization-online.org/DB_HTML/2018/10/6895.html
(note: there are therefore some notation changes from PySP1)
Note: we deviate from the paper's notation in the use of i and k
(i is used here as an arbitrary index, usually into the nonants at a node and
k is often used as the "key" (i.e., scenario name) for the local scenarios)
"""

class APH(ph_base.PHBase):  # ??????
    """
    Args:
        options (dict): PH options
        all_scenario_names (list): all scenario names
        scenario_creator (fct): returns a concrete model with special things
        scenario_denouement (fct): for post processing and reporting
        all_node_names (list of str): non-leaf node names
        scenario_creator_kwargs (dict): keyword arguments passed to
            `scenario_creator`.

    Attributes (partial list):
        local_scenarios (dict of scenario objects): concrete models with 
              extra data, key is name
        comms (dict): keys are node names values are comm objects.
        scenario_name_to_rank (dict): all scenario names
        local_scenario_names (list): names of locals 
        current_solver_options (dict): from options; callbacks might change
        synchronizer (object): asynch listener management
        scenario_creator_kwargs (dict): keyword arguments passed to
            `scenario_creator`.

    """
    def __init__(
        self,
        options,
        all_scenario_names,
        scenario_creator,
        scenario_denouement=None,
        all_nodenames=None,            
        mpicomm=None,
        scenario_creator_kwargs=None,
        extensions=None,
        extension_kwargs=None,
        PH_converger=None,
        rho_setter=None,
        variable_probability=None,
    ):
        super().__init__(
            options,
            all_scenario_names,
            scenario_creator,
            scenario_denouement,
            mpicomm=mpicomm,
            all_nodenames=all_nodenames,
            scenario_creator_kwargs=scenario_creator_kwargs,
            extensions=extensions,
            extension_kwargs=extension_kwargs,
            PH_converger=PH_converger,
            rho_setter=rho_setter,
            variable_probability=variable_probability,
        )

        self.phis = {} # phi values, indexed by scenario names
        self.tau_summand = 0  # place holder for iteration 1 reduce
        self.phi_summand = 0
        self.global_tau = 0
        self.global_phi = 0
        self.global_pusqnorm = 0 # ... may be out of date...
        self.global_pvsqnorm = 0
        self.global_pwsqnorm = 0
        self.global_pzsqnorm = 0
        self.local_pwsqnorm = 0
        self.local_pzsqnorm = 0
        self.conv = None
        self.use_lag = False if "APHuse_lag" not in options\
                        else options["APHuse_lag"]
        self.APHgamma = 1 if "APHgamma" not in options\
                        else options["APHgamma"]
        assert(self.APHgamma > 0)
        # TBD: use a property decorator for nu to enforce 0 < nu < 2
        self.nu = 1 # might be changed dynamically by an extension
        if "APHnu" in options:
            self.nu = options["APHnu"]
        assert 0 < self.nu and self.nu < 2
        self.dispatchrecord = dict()   # for local subproblems

    #============================
    def setup_Lens(self):
        """ We need to know the lengths of c-style vectors for listener_util
        """
        self.Lens = collections.OrderedDict({"FirstReduce": {},
                                            "SecondReduce": {}})

        for sname, scenario in self.local_scenarios.items():
            for node in scenario._mpisppy_node_list:
                self.Lens["FirstReduce"][node.name] \
                    = 3 * len(node.nonant_vardata_list)
                self.Lens["SecondReduce"][node.name] = 0 # only use root?
        self.Lens["FirstReduce"]["ROOT"] += self.n_proc  # for time of update
        # tau, phi, pusqnorm, pvsqnorm, pwsqnorm, pzsqnorm, secs
        self.Lens["SecondReduce"]["ROOT"] += 6 + self.n_proc 


    #============================
    def setup_dispatchrecord(self):
        # Start with a small number for iteration to randomize fist dispatch.
        for sname in self.local_subproblems:
            r = np.random.rand()
            self.dispatchrecord[sname] = [(r,0)]


    #============================
    def Update_y(self, dlist, verbose):
        # compute the new y (or set to zero if it is iter 1)
        # iter 1 is iter 0 post-solves when seen from the paper
        # dlist is used only after iter0 (it has the dispatched scen names)

        slist = [d[0] for d in dlist]  # just the names
        if self._PHIter != 1:
            for k,s in self.local_scenarios.items():
                if (not self.bundling and k in slist) \
                   or (self.bundling and s._mpisppy_data.bundlename in slist):
                    for (ndn,i), xvar in s._mpisppy_data.nonant_indices.items():
                        if not self.use_lag:
                            z_touse = s._mpisppy_model.z[(ndn,i)]._value
                            W_touse = pyo.value(s._mpisppy_model.W[(ndn,i)])
                        else:
                            z_touse = s._mpisppy_model.z_foropt[(ndn,i)]._value
                            W_touse = pyo.value(s._mpisppy_model.W_foropt[(ndn,i)])
                        # pyo.value vs. _value ??
                        s._mpisppy_model.y[(ndn,i)]._value = W_touse \
                                              + pyo.value(s._mpisppy_model.rho[(ndn,i)]) \
                                              * (xvar._value - z_touse)
                        if verbose and self.cylinder_rank == 0:
                            print ("node, scen, var, y", ndn, k,
                                   self.cylinder_rank, xvar.name,
                                   pyo.value(s._mpisppy_model.y[(ndn,i)]))
        else:
            for k,s in self.local_scenarios.items():
                for (ndn,i), xvar in s._mpisppy_data.nonant_indices.items():
                    s._mpisppy_model.y[(ndn,i)]._value = 0
            if verbose and self.cylinder_rank == 0:
                print ("All y=0 for iter1")


    #============================
    def compute_phis_summand(self):
        # update phis, return summand
        summand = 0.0
        for k,s in self.local_scenarios.items():
            self.phis[k] = 0.0
            for (ndn,i), xvar in s._mpisppy_data.nonant_indices.items():
                self.phis[k] += (pyo.value(s._mpisppy_model.z[(ndn,i)]) - xvar._value) \
                    *(pyo.value(s._mpisppy_model.W[(ndn,i)]) - pyo.value(s._mpisppy_model.y[(ndn,i)]))
            self.phis[k] *= pyo.value(s._mpisppy_probability)
            summand += self.phis[k]
        return summand

    #============================***********=========
    def listener_side_gig(self, synchro):
        """ Called by the listener after the first reduce.
        First, see if there are enough xbar contributions to proceed.
        If there are, then compute tau and phi.
        NOTE: it gets the synchronizer as an arg but self already has it.
        [WIP]
        We are going to disable the side_gig on self if we
        updated tau and phi.
        Massive side-effects: e.g., update xbar etc.

        Iter 1 (iter 0) in the paper is special: the v := u, which is a little
        complicated because we only compute y-bar.        

        """
        # This does unsafe things, so it can only be called when the worker is
        # in a tight loop that respects the data lock.

        verbose = self.options["verbose"]
        # See if we have enough xbars to proceed (need not be perfect)
        xbarin = 0 # count ranks (close enough to be a proxy for scenarios)
        self.synchronizer._unsafe_get_global_data("FirstReduce",
                                                  self.node_concats)
        self.synchronizer._unsafe_get_global_data("SecondReduce",
                                                  self.node_concats)
        # last_phi_tau_update_time
        lptut =  np.max(self.node_concats["SecondReduce"]["ROOT"][6:])
        logging.debug('enter side gig, last phi update={}'.format(lptut))
        for cr in range(self.n_proc):
            backdist = self.n_proc - cr
            logging.debug('*side_gig* cr {} on rank {} time {}'.\
                format(cr, self.cylinder_rank,
                    self.node_concats["FirstReduce"]["ROOT"][-backdist]))
            if  self.node_concats["FirstReduce"]["ROOT"][-backdist] \
                >= lptut:
                xbarin += 1
        if xbarin/self.n_proc < self.options["async_frac_needed"]:
            logging.debug('   not enough on rank {}'.format(self.cylinder_rank))
            # We have not really "done" the side gig.
            return

        # If we are still here, we have enough to do the calculations
        logging.debug('   good to go on rank {}'.format(self.cylinder_rank))
        if verbose and self.cylinder_rank == 0:
            print ("(%d)" % xbarin)
            
        # set the xbar, xsqbar, and ybar in all the scenarios
        for k,s in self.local_scenarios.items():
            nlens = s._mpisppy_data.nlens        
            for (ndn,i) in s._mpisppy_data.nonant_indices:
                s._mpisppy_model.xbars[(ndn,i)]._value \
                    = self.node_concats["FirstReduce"][ndn][i]
                s._mpisppy_model.xsqbars[(ndn,i)]._value \
                    = self.node_concats["FirstReduce"][ndn][nlens[ndn]+i]
                s._mpisppy_model.ybars[(ndn,i)]._value \
                    = self.node_concats["FirstReduce"][ndn][2*nlens[ndn]+i]

                if verbose and self.cylinder_rank == 0:
                    print ("rank, scen, node, var, xbar:",
                           self.cylinder_rank,k,ndn,s._mpisppy_data.nonant_indices[ndn,i].name,
                           pyo.value(s._mpisppy_model.xbars[(ndn,i)]))

        # There is one tau_summand for the rank; global_tau is out of date when
        # we get here because we could not compute it until the averages were.
        # vk is just going to be ybar directly
        if not hasattr(self, "uk"):
            self.uk = {} # indexed by sname and nonant index [sname][(ndn,i)]
        self.local_pusqnorm = 0  # local summand for probability weighted sqnorm
        self.local_pvsqnorm = 0
        new_tau_summand = 0  # for this rank
        for sname,s in self.local_scenarios.items():
            scen_usqnorm = 0.0
            scen_vsqnorm = 0.0
            if sname not in self.uk:
                self.uk[sname] = {}
            nlens = s._mpisppy_data.nlens        
            for (ndn,i), xvar in s._mpisppy_data.nonant_indices.items():
                self.uk[sname][(ndn,i)] = xvar._value \
                                          - pyo.value(s._mpisppy_model.xbars[(ndn,i)])
                # compute the usqnorm and vsqnorm (squared L2 norms)
                scen_usqnorm += self.uk[sname][(ndn,i)] \
                              * self.uk[sname][(ndn,i)]
                scen_vsqnorm += pyo.value(s._mpisppy_model.ybars[(ndn,i)]) \
                              * pyo.value(s._mpisppy_model.ybars[(ndn,i)])
            self.local_pusqnorm += pyo.value(s._mpisppy_probability) * scen_usqnorm
            self.local_pvsqnorm += pyo.value(s._mpisppy_probability) * scen_vsqnorm
            new_tau_summand += pyo.value(s._mpisppy_probability) \
                               * (scen_usqnorm + scen_vsqnorm/self.APHgamma)
                

            
        # tauk is the expecation of the sum sum of squares; update for this calc
        logging.debug('  in side-gig, old global_tau={}'.format(self.global_tau))
        logging.debug('  in side-gig, old summand={}'.format(self.tau_summand))
        logging.debug('  in side-gig, new summand={}'.format(new_tau_summand))
        self.global_tau = self.global_tau - self.tau_summand + new_tau_summand
        self.tau_summand = new_tau_summand # make available for the next reduce
        logging.debug('  in side-gig, new global_tau={}'.format(self.global_tau))

        # now we can get the local contribution to the phi_sum 
        if self.global_tau <= 0:
            logging.debug('  *** Negative tau={} on rank {}'\
                          .format(self.global_tau, self.cylinder_rank))
        self.phi_summand = self.compute_phis_summand()

        # prepare for the reduction that will take place after this side-gig
        self.local_concats["SecondReduce"]["ROOT"][0] = self.tau_summand
        self.local_concats["SecondReduce"]["ROOT"][1] = self.phi_summand
        self.local_concats["SecondReduce"]["ROOT"][2] = self.local_pusqnorm
        self.local_concats["SecondReduce"]["ROOT"][3] = self.local_pvsqnorm
        self.local_concats["SecondReduce"]["ROOT"][4] = self.local_pwsqnorm
        self.local_concats["SecondReduce"]["ROOT"][5] = self.local_pzsqnorm
        # we have updated our summands and the listener will do a reduction
        secs_so_far = time.perf_counter() - self.start_time
        # Put in a time only for this rank, so the "sum" is really a report
        self.local_concats["SecondReduce"]["ROOT"][6+self.cylinder_rank] = secs_so_far
        # This is run by the listener, so don't tell the worker you have done
        # it until you are sure you have.
        self.synchronizer._unsafe_put_local_data("SecondReduce",
                                                 self.local_concats)
        self.synchronizer.enable_side_gig = False  # we did it
        logging.debug(' exit side_gid on rank {}'.format(self.cylinder_rank))
        
    #============================
    def Compute_Averages(self, verbose=False):
        """Gather ybar, xbar and x squared bar for each node 
           and also distribute the values back to the scenarios.
           Compute the tau summand from self and distribute back tauk
           (tau_k is a scalar and special with respect to synchronizing).
           Compute the phi summand and reduce it.

        Args:
          verbose (boolean): verbose output

        note: this is a long routine because we need a reduce before
              we can do more calcs that need another reduce and I want
              to keep the reduce calls together.
        NOTE: see compute_xbar for more notes.
        note: DLW: think about multi-stage harder (March 2019); e.g. tau and phi

        """
        if not hasattr(self, "local_concats"):
            nodenames = [] # avoid repeated work
            self.local_concats = {"FirstReduce": {}, # keys are tree node names
                             "SecondReduce": {}}
            self.node_concats = {"FirstReduce": {}, # concat of xbar and xsqbar
                             "SecondReduce": {}} 

            # accumulate & concatenate all local contributions before the reduce

            # create the c-style storage for the concats
            for k,s in self.local_scenarios.items():
                nlens = s._mpisppy_data.nlens        
                for node in s._mpisppy_node_list:
                    if node.name not in nodenames:
                        ndn = node.name
                        nodenames.append(ndn)
                        mylen = self.Lens["FirstReduce"][ndn]
                        self.local_concats["FirstReduce"][ndn]\
                            = np.zeros(mylen, dtype='d')
                        self.node_concats["FirstReduce"][ndn]\
                            = np.zeros(mylen, dtype='d')
            # second reduce is tau and phi
            mylen = self.Lens["SecondReduce"]["ROOT"]
            self.local_concats["SecondReduce"]["ROOT"]\
                = np.zeros(mylen, dtype='d') 
            self.node_concats["SecondReduce"]["ROOT"]\
                = np.zeros(mylen, dtype='d')
        else: # concats are here, just zero them out. 
            """ delete this comment block after sept 2020:
            DLW Aug 2020: why zero?
            We zero them so we can use an accumulator in the next loop and
              that seems to be OK.
            """
            nodenames = []
            for k,s in self.local_scenarios.items():
                nlens = s._mpisppy_data.nlens        
                for node in s._mpisppy_node_list:
                    if node.name not in nodenames:
                        ndn = node.name
                        nodenames.append(ndn)
                        self.local_concats["FirstReduce"][ndn].fill(0)
                        self.node_concats["FirstReduce"][ndn].fill(0)                    
            self.local_concats["SecondReduce"]["ROOT"].fill(0)
            self.node_concats["SecondReduce"]["ROOT"].fill(0)

        # Compute the locals and concat them for the first reduce.
        # We don't need to lock here because the direct buffers are only accessed
        # by compute_global_data.
        for k,s in self.local_scenarios.items():
            nlens = s._mpisppy_data.nlens        
            for node in s._mpisppy_node_list:
                ndn = node.name
                for i in range(nlens[node.name]):
                    v_value = node.nonant_vardata_list[i]._value
                    self.local_concats["FirstReduce"][node.name][i] += \
                        (s._mpisppy_probability / node.uncond_prob) * v_value
                    logging.debug("  rank= {} scen={}, i={}, v_value={}".\
                                  format(global_rank, k, i, v_value))
                    self.local_concats["FirstReduce"][node.name][nlens[ndn]+i]\
                        += (s._mpisppy_probability / node.uncond_prob) * v_value * v_value
                    self.local_concats["FirstReduce"][node.name][2*nlens[ndn]+i]\
                        += (s._mpisppy_probability / node.uncond_prob) \
                           * pyo.value(s._mpisppy_model.y[(node.name,i)])

        # record the time
        secs_sofar = time.perf_counter() - self.start_time
        # only this rank puts a time for this rank, so the sum is a report
        self.local_concats["FirstReduce"]["ROOT"][3*nlens["ROOT"]+self.cylinder_rank] \
            = secs_sofar
        logging.debug('Compute_Averages at secs_sofar {} on rank {}'\
                      .format(secs_sofar, self.cylinder_rank))
                    
        self.synchronizer.compute_global_data(self.local_concats,
                                              self.node_concats,
                                              enable_side_gig = True,
                                              rednames = ["FirstReduce"],
                                              keep_up = True)
        # The lock is something to worry about here.
        while self.synchronizer.global_quitting == 0 \
              and self.synchronizer.enable_side_gig:
            # Other ranks could be reporting, so keep looking for them.
            self.synchronizer.compute_global_data(self.local_concats,
                                                  self.node_concats)
            if not self.synchronizer.enable_side_gig:
                logging.debug(' did side gig break on rank {}'.format(self.cylinder_rank))
                break
            else:
                logging.debug('   gig wait sleep on rank {}'.format(self.cylinder_rank))
                if verbose and self.cylinder_rank == 0:
                    print ('s'),
                time.sleep(self.options["async_sleep_secs"])

        # (if the listener still has the lock, compute_global_will wait for it)
        self.synchronizer.compute_global_data(self.local_concats,
                                              self.node_concats)
        # We  assign the global xbar, etc. as side-effect in the side gig, btw
        self.global_tau = self.node_concats["SecondReduce"]["ROOT"][0]
        self.global_phi = self.node_concats["SecondReduce"]["ROOT"][1]
        self.global_pusqnorm = self.node_concats["SecondReduce"]["ROOT"][2]
        self.global_pvsqnorm = self.node_concats["SecondReduce"]["ROOT"][3]
        self.global_pwsqnorm = self.node_concats["SecondReduce"]["ROOT"][4]
        self.global_pzsqnorm = self.node_concats["SecondReduce"]["ROOT"][5]

        logging.debug('Assigned global tau {} and phi {} on rank {}'\
                      .format(self.global_tau, self.global_phi, self.cylinder_rank))
 
    #============================
    def Update_theta_zw(self, verbose):
        """
        Compute and store theta, then update z and w and update
        the probability weighted norms.
        """
        if self.global_tau <= 0:
            logging.debug('|tau {}, rank {}'.format(self.global_tau, self.cylinder_rank))
            self.theta = 0   
        elif self.global_phi <= 0:
            logging.debug('|phi {}, rank {}'.format(self.global_phi, self.cylinder_rank))
            self.theta = 0
        else:
            self.theta = self.global_phi * self.nu / self.global_tau
        logging.debug('Iter {} assigned theta {} on rank {}'\
                      .format(self._PHIter, self.theta, self.cylinder_rank))

        oldpw = self.local_pwsqnorm
        oldpz = self.local_pzsqnorm
        self.local_pwsqnorm = 0
        self.local_pzsqnorm = 0
        # v is just ybar
        for k,s in self.local_scenarios.items():
            probs = pyo.value(s._mpisppy_probability)
            for (ndn, i) in s._mpisppy_data.nonant_indices:
                Wupdate = self.theta * self.uk[k][(ndn,i)]
                Ws = pyo.value(s._mpisppy_model.W[(ndn,i)]) + Wupdate
                s._mpisppy_model.W[(ndn,i)] = Ws 
                self.local_pwsqnorm += probs * Ws * Ws
                # iter 1 is iter 0 post-solves when seen from the paper
                if self._PHIter != 1:
                    zs = pyo.value(s._mpisppy_model.z[(ndn,i)])\
                     + self.theta * pyo.value(s._mpisppy_model.ybars[(ndn,i)])/self.APHgamma
                else:
                     zs = pyo.value(s._mpisppy_model.xbars[(ndn,i)])
                s._mpisppy_model.z[(ndn,i)] = zs 
                self.local_pzsqnorm += probs * zs * zs
                logging.debug("rank={}, scen={}, i={}, Ws={}, zs={}".\
                              format(global_rank, k, i, Ws, zs))
        # ? so they will be there next time? (we really need a third reduction)
        self.local_concats["SecondReduce"]["ROOT"][4] = self.local_pwsqnorm
        self.local_concats["SecondReduce"]["ROOT"][5] = self.local_pzsqnorm
        # The values we just computed can't be in the global yet, so update here
        self.global_pwsqnorm += (self.local_pwsqnorm - oldpw)
        self.global_pzsqnorm += (self.local_pzsqnorm - oldpz)
                
    #============================
    def Compute_Convergence(self, verbose=False):
        """
        The convergence metric is the sqrt of the sum of
        probability weighted unorm scaled by the probability weighted w norm
        probability weighted vnorm scaled by the probability weighted z norm
 
        Returns:
            update self.conv if appropriate
        """
        # dlw to dlw, April 2019: wnorm and znorm are in update_zw;
        # the u and v should be in the side gig.
        # you need a reduction on all the norms!!

        punorm = math.sqrt(self.global_pusqnorm)
        pwnorm = math.sqrt(self.global_pwsqnorm)
        pvnorm = math.sqrt(self.global_pvsqnorm)
        pznorm = math.sqrt(self.global_pzsqnorm)

        if pwnorm > 0 and pznorm > 0:
            self.conv = punorm / pwnorm + pvnorm / pznorm

        logging.debug('self.conv={} self.global_pusqnorm={} self.global_pwsqnorm={} self.global_pvsqnorm={} self.global_pzsqnorm={})'\
                      .format(self.conv, self.global_pusqnorm, self.global_pwsqnorm, self.global_pvsqnorm, self.global_pzsqnorm))
        # allow a PH converger, mainly for mpisspy to get xhat from a wheel conv
        if hasattr(self, "PH_conobject") and self.PH_convobject is not None:
            phc = self.PH_convobject(self, self.cylinder_rank, self.n_proc)
            logging.debug("PH converger called (returned {})".format(phc))


    #==========
    def _update_foropt(self, dlist):
        # dlist is a list of subproblem names that were dispatched
        assert self.use_lag
        """
        if not self.bundling:
            phidict = self.phis
        else:
            phidict = {k: self.phis[self.local_subproblems[k].scen_list[0]]}
        """
        if not self.bundling:
            for dl in dlist:
                scenario = self.local_scenarios[dl[0]]
                for (ndn,i), xvar in scenario._mpisppy_data.nonant_indices.items():
                    scenario._mpisppy_model.z_foropt[(ndn,i)] = scenario._mpisppy_model.z[(ndn,i)]
                    scenario._mpisppy_model.W_foropt[(ndn,i)] = scenario._mpisppy_model.W[(ndn,i)]
        else:
            for dl in dlist:
                for sname in self.local_subproblems[dl[0]].scen_list:
                    scenario = self.local_scenarios[sname]
                    for (ndn,i), xvar in scenario._mpisppy_data.nonant_indices.items():
                        scenario._mpisppy_model.z_foropt[(ndn,i)] = scenario._mpisppy_model.z[(ndn,i)]
                        scenario._mpisppy_model.W_foropt[(ndn,i)] = scenario._mpisppy_model.W[(ndn,i)]


    #====================================================================
    def APH_solve_loop(self, solver_options=None,
                       use_scenarios_not_subproblems=False,
                       dtiming=False,
                       gripe=False,
                       disable_pyomo_signal_handling=False,
                       tee=False,
                       verbose=False,
                       dispatch_frac=1):
        """See phbase.solve_loop. Loop over self.local_subproblems and solve
            them in a manner dicated by the arguments. In addition to
            changing the Var values in the scenarios, update
            _PySP_feas_indictor for each.

        Args:
            solver_options (dict or None): the scenario solver options
            use_scenarios_not_subproblems (boolean): for use by bounds
            dtiming (boolean): indicates that timing should be reported
            gripe (boolean): output a message if a solve fails
            disable_pyomo_signal_handling (boolean): set to true for asynch, 
                                                     ignored for persistent solvers.
            tee (boolean): show solver output to screen if possible
            verbose (boolean): indicates verbose output
            dispatch_frac (float): fraction to send out for solution based on phi

        Returns:
            dlist (list of (str, float): (dispatched name, phi )
        """
        #==========
        def _vb(msg): 
            if verbose and self.cylinder_rank == 0:
                print ("(rank0) " + msg)
        _vb("Entering solve_loop function.")


        #==========
        def _best_phis():
            # for dispatch based on phi
            # note that when there is no bundling, scenarios are subproblems
            # {k: v for k, v in sorted(x.items(), key=lambda item: item[1])}
            if use_scenarios_not_subproblems:
                s_source = self.local_scenarios
                phidict = self.phis
            else:
                s_source = self.local_subproblems
                if not self.bundling:
                    phidict = self.phis
                else:
                    phidict = {k: self.phis[self.local_subproblems[k].scen_list[0]] for k in s_source.keys()}
            # dict(sorted(phidict.items(), key=lambda item: item[1]))
            sortedbyphi = {k: v for k, v in sorted(phidict.items(), key=lambda item: item[1])}

            return s_source, sortedbyphi


        #========
        def _dispatch_list(scnt):
            # Return the entire source dict and list of scnt (subproblems,phi) 
            # pairs for dispatch.
            retval = list()  # the list to return
            s_source, sortedbyphi = _best_phis()
            i = 0
            for k,p in sortedbyphi.items():
                if p < 0:
                    retval.append((k,p))
                    i += 1
                    if i >= scnt:
                        logging.debug("Dispatch list w/neg phi after {}/{} (frac needed={})".\
                                      format(i, len(sortedbyphi), dispatch_frac))
                        return s_source, retval

            # If we are still here, there were not enough w/negative phi values.
            if i == 0 and self.nu == 1.0 and self._PHIter > 1:
                print(f"WARNING: no negative phi on rank {self.cylinder_rank}"
                      f" at iteration {self._PHIter}")
            # Use phi as  tie-breaker (sort by the most recent dispatch tuple)
            sortedbyI = {k: v for k, v in sorted(self.dispatchrecord.items(), 
                                                 key=lambda item: item[1][-1])}
            for k,t in sortedbyI.items():
                if k in retval:
                    continue
                retval.append((k, sortedbyphi[k]))  # sname, phi
                i += 1
                if i >= scnt:
                    logging.debug("Dispatch list complete after {}/{} (frac needed={})".\
                                  format(i, len(sortedbyphi), dispatch_frac))
                    break
            return s_source, retval


        # body of fct starts hare
        logging.debug("  early APH solve_loop for rank={}".format(self.cylinder_rank))

        scnt = max(1, len(self.dispatchrecord) * dispatch_frac)
        s_source, dlist = _dispatch_list(scnt)
        for dguy in dlist:
            k = dguy[0]   # name of who to dispatch
            p = dguy[1]   # phi
            s = s_source[k]
            self.dispatchrecord[k].append((self._PHIter, p))
            logging.debug("  in APH solve_loop rank={}, k={}, phi={}".\
                          format(self.cylinder_rank, k, p))
            pyomo_solve_time = self.solve_one(solver_options, k, s,
                                              dtiming=dtiming,
                                              verbose=verbose,
                                              tee=tee,
                                              gripe=gripe,
                disable_pyomo_signal_handling=disable_pyomo_signal_handling
            )

        if dtiming:
            all_pyomo_solve_times = self.mpicomm.gather(pyomo_solve_time, root=0)
            if self.cylinder_rank == 0:
                print("Pyomo solve times (seconds):")
                print("\tmin=%4.2f mean=%4.2f max=%4.2f" %
                      (np.min(all_pyomo_solve_times),
                      np.mean(all_pyomo_solve_times),
                      np.max(all_pyomo_solve_times)))
        return dlist

    #========
    def _print_conv_detail(self):
        print("Convergence Metric=",self.conv)
        punorm = math.sqrt(self.global_pusqnorm)
        pwnorm = math.sqrt(self.global_pwsqnorm)
        pvnorm = math.sqrt(self.global_pvsqnorm)
        pznorm = math.sqrt(self.global_pzsqnorm)
        print(f'   punorm={punorm} pwnorm={pwnorm} pvnorm={pvnorm} pznorm={pznorm}')
        if pwnorm > 0 and pznorm > 0:
            print(f"    scaled U term={punorm / pwnorm}; scaled V term={pvnorm / pznorm}")
        else:
            print(f"    ! convergence metric cannot be computed due to zero-divide")


    #========
    def display_details(self, msg):
        """Ouput as much as you can about the current state"""
        print(f"hello {msg}")
        print(f"*** global rank {global_rank} display details: {msg}")
        print(f"zero-based iteration number {self._PHIter}")
        self._print_conv_detail()
        print(f"phi={self.global_phi}, nu={self.nu}, tau={self.global_tau} so theta={self.theta}")
        print(f"{'Nonants for':19} {'x':8} {'z':8} {'W':8} {'u':8} ")
        for k,s in self.local_scenarios.items():
            print(f"   Scenario {k}")
            for (ndn,i), xvar in s._mpisppy_data.nonant_indices.items():
                print(f"   {(ndn,i)} {xvar._value:9.3} "
                      f"{s._mpisppy_model.z[(ndn,i)]._value:9.3}"
                      f"{s._mpisppy_model.W[(ndn,i)]._value:9.3}"
                      f"{self.uk[k][(ndn,i)]:9.3}")
      

    #====================================================================
    def APH_iterk(self, spcomm):
        """ Loop for the main iterations (called by synchronizer).

        Args:
        spcomm (SPCommunitator object): to communicate intra and inter

        Updates: 
            self.conv (): APH convergence

        """
        logging.debug('==== enter iterk on rank {}'.format(self.cylinder_rank))
        verbose = self.options["verbose"]
        have_extensions = self.extensions is not None

        # We have the "bottom of the loop at the top"
        # so we need a dlist to get the ball rolling (it might not be used)
        dlist = [(sn, 0.0) for sn in self.local_scenario_names]
        
        # put dispatch_frac on the object so extensions can modify it
        self.dispatch_frac = self.options["dispatch_frac"]\
                             if "dispatch_frac" in self.options else 1

        have_converger = self.PH_converger is not None
        dprogress = self.options["display_progress"]
        dtiming = self.options["display_timing"]
        ddetail = "display_convergence_detail" in self.options and\
            self.options["display_convergence_detail"]
        self.conv = None
        # The notion of an iteration is unclear
        # we enter after the iteration 0 solves, so do updates first
        for self._PHIter in range(1, self.options["PHIterLimit"]+1):
            if self.synchronizer.global_quitting:
                break
            iteration_start_time = time.time()

            if dprogress and self.cylinder_rank == 0:
                print("")
                print ("Initiating APH Iteration",self._PHIter)
                print("")

            self.Update_y(dlist, verbose)
            # Compute xbar, etc
            logging.debug('pre Compute_Averages on rank {}'.format(self.cylinder_rank))
            self.Compute_Averages(verbose)
            logging.debug('post Compute_Averages on rank {}'.format(self.cylinder_rank))
            if self.global_tau <= 0:
                logging.debug('***tau is 0 on rank {}'.format(self.cylinder_rank))

            # Apr 2019 dlw: If you want the convergence crit. to be up to date,
            # do this as a listener side-gig and add another reduction.
            self.Update_theta_zw(verbose)
            self.Compute_Convergence()  # updates conv
            phisum = self.compute_phis_summand() # post-step phis for dispatch
            logging.debug('phisum={} after step on {}'.format(phisum, self.cylinder_rank))

            # ORed checks for convergence
            if spcomm is not None and type(spcomm) is not mpi4py.MPI.Intracomm:
                spcomm.sync_with_spokes()
                logging.debug('post sync_with_spokes on rank {}'.format(self.cylinder_rank))
                if spcomm.is_converged():
                    break    
            if have_converger:
                if self.convobject.is_converged():
                    converged = True
                    if self.cylinder_rank == 0:
                        print("User-supplied converger determined termination criterion reached")
                    break
            if ddetail:
                self.display_details("pre-solve loop (everything is updated from prev iter)")
            # slight divergence from PH, where mid-iter is before conv
            if have_extensions:
                self.extobject.miditer()
            
            teeme = ("tee-rank0-solves" in self.options) \
                 and (self.options["tee-rank0-solves"] == True
                      and self.cylinder_rank == 0)
            # Let the solve loop deal with persistent solvers & signal handling
            # Aug2020 switch to a partial loop xxxxx maybe that is enough.....
            # Aug2020 ... at least you would get dispatch
            if self._PHIter == 1:
                savefrac = self.dispatch_frac
                self.dispatch_frac = 1   # to get a decent w for everyone
            logging.debug('pre APH_solve_loop on rank {}'.format(self.cylinder_rank))
            dlist = self.APH_solve_loop(solver_options = \
                                        self.current_solver_options,
                                        dtiming=dtiming,
                                        gripe=True,
                                        disable_pyomo_signal_handling=True,
                                        tee=teeme,
                                        verbose=verbose,
                                        dispatch_frac=self.dispatch_frac)

            logging.debug('post APH_solve_loop on rank {}'.format(self.cylinder_rank))
            if self._PHIter == 1:
                 self.dispatch_frac = savefrac
            if have_extensions:
                self.extobject.enditer()

            if dprogress and self.cylinder_rank == 0:
                print("")
                print("After APH Iteration",self._PHIter)
                if not ddetail:
                    self._print_conv_detail()
                print("Iteration time: %6.2f" \
                      % (time.time() - iteration_start_time))
                print("Elapsed time:   %6.2f" \
                      % (time.perf_counter() - self.start_time))
            if self.use_lag:
                self._update_foropt(dlist)

        logging.debug('Setting synchronizer.quitting on rank %d' % self.cylinder_rank)
        self.synchronizer.quitting = 1

    #====================================================================
    def APH_main(self, spcomm=None, finalize=True):

        """Execute the APH algorithm.
        Args:
            spcomm (SPCommunitator object): for intra or inter communications
            finalize (bool, optional, default=True):
                        If True, call self.post_loops(), if False, do not,
                        and return None for Eobj

        Returns:
            conv, Eobj, trivial_bound: 
                        The first two CANNOT BE EASILY INTERPRETED. 
                        Eobj is the expected,  weighted objective with 
                        proximal term. It is not directly useful.
                        The trivial bound is computed after iter 0
        NOTE:
            You need an xhat finder either in denoument or in an extension.
        """
        # Prep needs to be before iter 0 for bundling
        # (It could be split up)
        self.PH_Prep(attach_duals=False, attach_prox=False)

        # Begin APH-specific Prep
        for sname, scenario in self.local_scenarios.items():    
            # ys is plural of y
            scenario._mpisppy_model.y = pyo.Param(scenario._mpisppy_data.nonant_indices.keys(),
                                     initialize = 0.0,
                                     mutable = True)
            scenario._mpisppy_model.ybars = pyo.Param(scenario._mpisppy_data.nonant_indices.keys(),
                                        initialize = 0.0,
                                        mutable = True)
            scenario._mpisppy_model.z = pyo.Param(scenario._mpisppy_data.nonant_indices.keys(),
                                     initialize = 0.0,
                                     mutable = True)
            # lag: we will support lagging back only to the last solve
            # IMPORTANT: pyomo does not support a second reference so no:
            # scenario._mpisppy_model.z_foropt = scenario._mpisppy_model.z
            
            if self.use_lag:
                scenario._mpisppy_model.z_foropt = pyo.Param(scenario._mpisppy_data.nonant_indices.keys(),
                                         initialize = 0.0,
                                         mutable = True)
                scenario._mpisppy_model.W_foropt = pyo.Param(scenario._mpisppy_data.nonant_indices.keys(),
                                         initialize = 0.0,
                                         mutable = True)
                
            objfct = find_active_objective(scenario)
                
            if self.use_lag:
                for (ndn,i), xvar in scenario._mpisppy_data.nonant_indices.items():
                    # proximal term
                    objfct.expr +=  scenario._mpisppy_model.prox_on * \
                        (scenario._mpisppy_model.rho[(ndn,i)] /2.0) * \
                        (xvar**2 - 2.0*xvar*scenario._mpisppy_model.z_foropt[(ndn,i)] + scenario._mpisppy_model.z_foropt[(ndn,i)]**2)                                            
                    # W term
                    objfct.expr +=  scenario._mpisppy_model.W_on * scenario._mpisppy_model.W_foropt[ndn,i] * xvar
            else:
                for (ndn,i), xvar in scenario._mpisppy_data.nonant_indices.items():
                    # proximal term
                    objfct.expr +=  scenario._mpisppy_model.prox_on * \
                        (scenario._mpisppy_model.rho[(ndn,i)] /2.0) * \
                        (xvar**2 - 2.0*xvar*scenario._mpisppy_model.z[(ndn,i)] + scenario._mpisppy_model.z[(ndn,i)]**2)                        
                    # W term
                    objfct.expr +=  scenario._mpisppy_model.W_on * scenario._mpisppy_model.W[ndn,i] * xvar

        # End APH-specific Prep
        
        self.subproblem_creation(self.options["verbose"])

        trivial_bound = self.Iter0()

        self.setup_Lens()
        self.setup_dispatchrecord()

        sleep_secs = self.options["async_sleep_secs"]

        lkwargs = None  # nothing beyond synchro
        listener_gigs = {"FirstReduce": (self.listener_side_gig, lkwargs),
                         "SecondReduce": None}
        self.synchronizer = listener_util.Synchronizer(comms = self.comms,
                                                    Lens = self.Lens,
                                                    work_fct = self.APH_iterk,
                                                    rank = self.cylinder_rank,
                                                    sleep_secs = sleep_secs,
                                                    asynch = True,
                                                    listener_gigs = listener_gigs)
        args = [spcomm] if spcomm is not None else [fullcomm]
        kwargs = None  # {"extensions": extensions}
        self.synchronizer.run(args, kwargs)

        if finalize:
            Eobj = self.post_loops()
        else:
            Eobj = None

#        print(f"Debug: here's the dispatch record for rank={self.global_rank}")
#        for k,v in self.dispatchrecord.items():
#            print(k, v)
#            print()
#        print("End dispatch record")

        return self.conv, Eobj, trivial_bound

#************************************************************
if __name__ == "__main__":
    # hardwired by dlw for debugging
    import mpisppy.tests.examples.farmer as refmodel

    PHopt = {}
    PHopt["asynchronousPH"] = False # APH is *projective* and always APH
    PHopt["solvername"] = "cplex"
    PHopt["PHIterLimit"] = 5
    PHopt["defaultPHrho"] = 1
    PHopt["APHgamma"] = 1
    PHopt["convthresh"] = 0.001
    PHopt["verbose"] = True
    PHopt["display_timing"] = True
    PHopt["display_progress"] = True
    # one way to set up options (never mind that this is not a MIP)
    PHopt["iter0_solver_options"] = sputils.option_string_to_dict("mipgap=0.01")
    # another way
    PHopt["iterk_solver_options"] = {"mipgap": 0.001}

    ScenCount = 3
    scenario_creator_kwargs = {'use_integer': False, "crops_multiplier": 1}
    all_scenario_names = list()
    for sn in range(ScenCount):
        all_scenario_names.append("scen"+str(sn))
    # end hardwire

    scenario_creator = refmodel.scenario_creator
    scenario_denouement = refmodel.scenario_denouement

    PHopt["async_frac_needed"] = 0.5
    PHopt["async_sleep_secs"] = 0.5
    aph = APH(
        PHopt,
        all_scenario_names,
        scenario_creator,
        scenario_denouement,
        scenario_creator_kwargs=scenario_creator_kwargs,
    )


    """
    import xhatlooper
    PHopt["xhat_looper_options"] =  {"xhat_solver_options":\
                                     PHopt["iterk_solver_options"],
                                     "scen_limit": 3,
                                     "csvname": "looper.csv"}
    """
    conv, obj, bnd = aph.APH_main()

    if aph.cylinder_rank == 0:
        print ("E[obj] for converged solution (probably NOT non-anticipative)",
               obj)

    dopts = sputils.option_string_to_dict("mipgap=0.001")
    objbound = aph.post_solve_bound(solver_options=dopts, verbose=False)
    if (aph.cylinder_rank == 0):
        print ("**** Lagrangian objective function bound=",objbound)
        print ("(probably converged way too early, BTW)")
